"""MEModel class"""

"""
Copyright 2024, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from bluepyemodel.emodel_pipeline.emodel import EModelMixin


class MEModel(EModelMixin):
    """Combination of an EModel and a Morphology. Should contain ids of these resources,
    as well as analysis plotting figure paths."""

    def __init__(
        self,
        seed=None,
        emodel_metadata=None,
        emodel_id=None,
        morphology_id=None,
        validated=False,
        status="initialized",
    ):
        """Init

        Args:
            seed (int): seed used during optimisation for this emodel.
            emodel_metadata (EModelMetadata): metadata of the model (emodel name, etype, ttype, ...)
            emodel_id (str): nexus if of the e-model used in this me-model
            morphology_id (str): nexus id of the morphology used in this me-model
            validated (bool): whether the MEModel has been validated by user
            status (str): whether the analysis has run or not. Can be "initialized" or "done".
        """

        self.emodel_metadata = emodel_metadata
        self.seed = seed

        self.emodel_id = emodel_id
        self.morphology_id = morphology_id

        self.validated = validated
        self.status = status

    def get_related_nexus_ids(self):
        uses = []
        if self.emodel_id:
            uses.append({"id": self.emodel_id, "type": "EModel"})
        if self.morphology_id:
            uses.append({"id": self.morphology_id, "type": "NeuronMorphology"})
        return {"uses": uses}

    def as_dict(self):
        pdf_dependencies = self.build_pdf_dependencies(self.seed)

        return {
            "nexus_images": pdf_dependencies,
            "seed": self.seed,
            "emodel_id": self.emodel_id,
            "morphology_id": self.morphology_id,
            "validated": self.validated,
            "status": self.status,
        }



"""Comb stimulus class."""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

import numpy

from bluepyemodel.ecode.stimulus import BPEM_stimulus

logger = logging.getLogger(__name__)


class Comb(BPEM_stimulus):
    # pylint: disable=line-too-long,anomalous-backslash-in-string

    """Comb current stimulus which consists of regularly spaced short steps, aimed at
    generating a train of spikes.

    .. code-block:: none

              holdi         amp       holdi        amp       holdi           .   .   .
                :            :          :           :          :
                :       ___________     :      ___________     :     ___                  _____
                :      |           |    :     |           |    :    |                          |
                :      |           |    :     |           |    :    |        * n_steps         |
                :      |           |    :     |           |    :    |        .   .   .         |
                :      |           |    :     |           |    :    |                          |
        |______________|           |__________|           |_________|                          |_____
        :              :           :          :           :         :                                ^
        :              :           :          :           :         :                                :
        :              :           :          :           :         :                                :
         <--  delay  --><-duration->           <-duration->         :        .   .   .     totduration
                        <--   inter_delay  --><--  inter_delay   -->

    """

    name = "Comb"

    def __init__(self, location, **kwargs):
        """Constructor
        Args:
            location(Location): location of stimulus
            inter_delay (float): time between each step beginnings in ms
            n_steps (int): number of steps for the stimulus
            amp (float): amplitude of each step(nA)
            delay (float): time at which the first current spike begins (ms)
            duration (float): duration of each step (ms)
            totduration (float): total duration of the whole stimulus (ms)
        """
        self.inter_delay = kwargs.get("inter_delay", 5)
        self.n_steps = kwargs.get("n_steps", 20)
        self.amp = kwargs.get("amp", 40)
        self.delay = kwargs.get("delay", 200.0)
        self.duration = kwargs.get("duration", 0.5)
        self.total_duration = kwargs.get("totduration", 350.0)
        self.holding = 0.0  # hardcoded holding for now (holding_current is modified externally)

        if self.stim_end > self.total_duration:
            raise ValueError(
                "stim_end is larger than total_duration: {self.stim_end} > {self.total_duration})"
            )

        super().__init__(
            location=location,
        )

    @property
    def stim_start(self):
        return self.delay

    @property
    def stim_end(self):
        return self.delay + self.n_steps * self.duration

    @property
    def amplitude(self):
        return self.amp

    def instantiate(self, sim=None, icell=None):
        """Run stimulus"""

        icomp = self.location.instantiate(sim=sim, icell=icell)

        self.iclamp = sim.neuron.h.IClamp(icomp.x, sec=icomp.sec)
        self.iclamp.dur = self.total_duration

        self.current_vec = sim.neuron.h.Vector()
        self.time_vec = sim.neuron.h.Vector()
        self.time_vec.append(self.holding)
        self.current_vec.append(self.holding)

        for step in range(self.n_steps):
            _delay = step * self.inter_delay + self.delay
            self.time_vec.append(_delay)
            self.current_vec.append(self.holding)

            self.time_vec.append(_delay)
            self.current_vec.append(self.amplitude)

            self.time_vec.append(_delay + self.duration)
            self.current_vec.append(self.amplitude)

            self.time_vec.append(_delay + self.duration)
            self.current_vec.append(self.holding)

        self.time_vec.append(self.total_duration)
        self.current_vec.append(self.holding)

        self.iclamp.delay = 0
        self.current_vec.play(
            self.iclamp._ref_amp,  # pylint:disable=W0212
            self.time_vec,
            1,
            sec=icomp.sec,
        )

    def generate(self, dt=0.1):
        """Return current time series"""

        t = numpy.arange(0.0, self.total_duration, dt)
        current = numpy.full(t.shape, self.holding, dtype="float64")

        for step in range(self.n_steps):
            _delay = step * self.inter_delay + self.delay
            ton_idx = int(_delay / dt)
            toff_idx = int((_delay + self.duration) / dt)
            current[ton_idx:toff_idx] = self.amplitude

        return t, current



""" Emodels evaluation functions """

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import copy
import glob
import logging
from pathlib import Path

import numpy
from bluepyopt.ephys.responses import TimeVoltageResponse

from bluepyemodel.access_point import get_access_point
from bluepyemodel.access_point.local import LocalAccessPoint
from bluepyemodel.evaluation.evaluator import create_evaluator
from bluepyemodel.model import model
from bluepyemodel.tools.mechanisms import compile_mechs_in_emodel_dir
from bluepyemodel.tools.mechanisms import delete_compiled_mechanisms
from bluepyemodel.tools.utils import make_dir

logger = logging.getLogger(__name__)


def locally_store_responses(emodel):
    """Locally store the responses.

    Arguments:
        emodel (EModel): the emodel which responses are to be stored
    """
    output_dir = f"./recordings/{emodel.emodel_metadata.as_string(emodel.seed)}"
    make_dir(output_dir)
    for key, resp in emodel.responses.items():
        output_path = Path(output_dir) / ".".join((key, "dat"))
        if not ("holding_current" in key or "threshold_current" in key or "bpo" in key):
            if resp["time"] is not None and resp["voltage"] is not None:
                time = numpy.array(resp["time"])
                data = numpy.array(resp["voltage"])  # even current will be named voltage here

                numpy.savetxt(output_path, numpy.transpose(numpy.vstack((time, data))))
        else:
            numpy.savetxt(output_path, numpy.asarray([resp]))


def check_local_responses_presence(emodels, cell_eval):
    """Returns True if there is a local response file for each emodel.

    Arguments:
        emodels (list of EModel): the emodel that need a response to be checked.
        cell_eval (CellEvaluator): evaluator for the cell model/protocols/e-feature set.
    """
    for emodel in emodels:
        output_dir = f"./recordings/{emodel.emodel_metadata.as_string(emodel.seed)}"
        if not Path(output_dir).is_dir():
            return False

        # only check voltage, since some currents are not recorded
        # because non existent depending on location
        if not all(
            (
                (Path(output_dir) / ".".join((rec.name, "dat"))).is_file()
                for prot in cell_eval.fitness_protocols["main_protocol"].protocols.values()
                for rec in prot.recordings
                if rec.variable == "v"
            )
        ):
            return False

    return True


def load_responses_from_local_files(emodels, cell_eval):
    """Returns responses from locally stored files.

    Arguments:
        emodels (list of EModel): the emodel that need a response to be checked.
        cell_eval (CellEvaluator): evaluator for the cell model/protocols/e-feature set.
    """
    responses_list = []
    for emodel in emodels:
        responses = {}
        output_dir = f"./recordings/{emodel.emodel_metadata.as_string(emodel.seed)}"

        for filepath in glob.glob(str(Path(output_dir) / "*.dat")):
            response_key = Path(filepath).stem
            response_name = response_key.split(".")[0]
            data = numpy.loadtxt(filepath)
            if not (
                "holding_current" in response_key
                or "threshold_current" in response_key
                or "bpo" in response_key
            ):
                responses[response_key] = TimeVoltageResponse(
                    name=response_name, time=data[:, 0], voltage=data[:, 1]
                )
            else:
                responses[response_key] = data
        responses["evaluator"] = copy.deepcopy(cell_eval)

        responses_list.append(responses)

    return responses_list


def get_responses(to_run):
    """Compute the voltage responses of a set of parameters.

    Args:
        to_run (dict): of the form
            to_run = {"evaluator": CellEvaluator, "parameters": Dict}
    """

    eva = to_run["evaluator"]
    params = to_run["parameters"]

    eva.cell_model.unfreeze(params)

    responses = eva.run_protocols(protocols=eva.fitness_protocols.values(), param_values=params)
    responses["evaluator"] = eva

    return responses


def compute_responses(
    access_point,
    cell_evaluator,
    map_function,
    seeds=None,
    preselect_for_validation=False,
    store_responses=False,
    load_from_local=False,
):
    """Compute the responses of the emodel to the optimisation and validation protocols.

    Args:
        access_point (DataAccessPoint): API used to access the data.
        emodel (str): name of the emodel. Has to match the name of the emodel
            under which the configuration data are stored.
        cell_evaluator (CellEvaluator): evaluator for the cell model/protocols/e-feature set.
        map_function (map): used to parallelize the evaluation of the
            individual in the population.
        seeds (list): if not None, filter emodels to keep only the ones with these seeds.
        preselect_for_validation (bool): if True,
            only select models that have not been through validation yet.
        store_responses (bool): whether to locally store the responses.
        load_from_local (bool): True to load responses from locally saved recordings.
    Returns:
        emodels (list): list of emodels.
    """

    emodels = access_point.get_emodels()

    if seeds:
        emodels = [model for model in emodels if model.seed in seeds]
    if access_point.emodel_metadata.iteration:
        emodels = [
            model
            for model in emodels
            if model.emodel_metadata.iteration == access_point.emodel_metadata.iteration
        ]
    if preselect_for_validation:
        emodels = [model for model in emodels if model.passed_validation is None]

    if emodels:
        logger.info("In compute_responses, %s emodels found to evaluate.", len(emodels))

        to_run = []
        for mo in emodels:
            to_run.append(
                {
                    "evaluator": copy.deepcopy(cell_evaluator),
                    "parameters": mo.parameters,
                }
            )

        if load_from_local and check_local_responses_presence(emodels, cell_evaluator):
            logger.info(
                "Local responses file found. Loading them from files instead of recomputing them"
            )
            responses = load_responses_from_local_files(emodels, cell_evaluator)
        else:
            responses = list(map_function(get_responses, to_run))

        for mo, r in zip(emodels, responses):
            mo.responses = r
            mo.evaluator = r.pop("evaluator")
            if store_responses and not load_from_local:
                locally_store_responses(mo)

    else:
        logger.warning(
            "In compute_responses, no emodel for %s",
            access_point.emodel_metadata.emodel,
        )

    return emodels


def fill_initial_parameters(evaluator, initial_parameters):
    """Freezes the parameters of the evaluator that are present in the informed parameter set."""
    # pylint: disable=protected-access
    replaced = []

    for p in evaluator.cell_model.params:
        if (
            p in initial_parameters
            and evaluator.cell_model.params[p].bounds is None
            and evaluator.cell_model.params[p]._value is None
        ):
            logger.info(
                "Parameter %s is set to its value from previous emodel: %s",
                evaluator.cell_model.params[p].name,
                initial_parameters[p],
            )
            evaluator.cell_model.params[p]._value = initial_parameters[p]
            evaluator.cell_model.params[p].frozen = True
            replaced.append(evaluator.cell_model.params[p].name)

    evaluator.params = [p for p in evaluator.params if p.name not in replaced]
    evaluator.param_names = [pn for pn in evaluator.param_names if pn not in replaced]


def get_evaluator_from_access_point(
    access_point,
    stochasticity=None,
    include_validation_protocols=False,
    timeout=None,
    use_fixed_dt_recordings=False,
    record_ions_and_currents=False,
):
    """Create an evaluator for the emodel.

    Args:
        access_point (DataAccessPoint): API used to access the database
        stochasticity (bool): should channels behave stochastically if they can.
        include_validation_protocols (bool): should the validation protocols
            and validation efeatures be added to the evaluator.
        timeout (float): duration (in second) after which the evaluation of a
            protocol will be interrupted.
        use_fixed_dt_recordings (bool): used for legacy currentscape
            to record at a fixed dt of 0.1 ms.
        record_ions_and_currents (bool): whether to add the ion and non-specific currents
            and the ionic concentration to the recordings.

    Returns:
        bluepyopt.ephys.evaluators.CellEvaluator
    """

    model_configuration = access_point.get_model_configuration()
    fitness_calculator_configuration = access_point.get_fitness_calculator_configuration(
        record_ions_and_currents=record_ions_and_currents
    )

    if model_configuration.morph_modifiers:
        morph_modifiers = model_configuration.morph_modifiers
    elif access_point.pipeline_settings:
        morph_modifiers = access_point.pipeline_settings.morph_modifiers
    else:
        morph_modifiers = None

    cell_model = model.create_cell_model(
        name=access_point.emodel_metadata.emodel,
        model_configuration=model_configuration,
        morph_modifiers=morph_modifiers,
    )

    mechanisms_directory = access_point.get_mechanisms_directory()
    if isinstance(access_point, LocalAccessPoint):
        if (
            Path.cwd() != access_point.emodel_dir.resolve()
            and access_point.emodel_metadata.iteration
        ):
            delete_compiled_mechanisms()
            if not (access_point.emodel_dir / "x86_64" / "special").is_file():
                compile_mechs_in_emodel_dir(mechanisms_directory)
        else:
            # if x86_64 present in main repo AND mechanisms_directory given to simulator
            # NEURON loads mechanisms twice and crash
            mechanisms_directory = None

    evaluator = create_evaluator(
        cell_model=cell_model,
        fitness_calculator_configuration=fitness_calculator_configuration,
        pipeline_settings=access_point.pipeline_settings,
        stochasticity=stochasticity,
        timeout=timeout,
        include_validation_protocols=include_validation_protocols,
        mechanisms_directory=mechanisms_directory,
        use_fixed_dt_recordings=use_fixed_dt_recordings,
    )

    start_from_emodel = access_point.pipeline_settings.start_from_emodel

    if start_from_emodel is not None:
        access_point_type = "local" if isinstance(access_point, LocalAccessPoint) else "nexus"

        seed = start_from_emodel.pop("seed", None)

        if access_point_type == "local":
            kwargs = {
                "emodel_dir": access_point.emodel_dir,
                "final_path": access_point.final_path,
                "recipes_path": access_point.recipes_path,
            }
        else:
            raise NotImplementedError("start_from_emodel not implemented for Nexus access point")

        starting_access_point = get_access_point(
            access_point=access_point_type, **start_from_emodel, **kwargs
        )

        emodels = starting_access_point.get_emodels()
        if not emodels:
            raise ValueError(
                f"Cannot start optimisation of {access_point.emodel_metadata.emodel} because"
                f" there are no emodels for {start_from_emodel}"
            )

        if seed is None:
            initial_parameters = sorted(emodels, key=lambda x: x.fitness)[0].parameters
        else:
            initial_parameters = next(
                (e.parameters for e in emodels if str(e.seed) == str(seed)), None
            )
            if initial_parameters is None:
                raise ValueError(
                    f"Cannot start optimisation of {access_point.emodel_metadata.emodel} because"
                    f" there are no emodels for {start_from_emodel}"
                )

        fill_initial_parameters(evaluator, initial_parameters)

    return evaluator



"""Optimisation function"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import os
from pathlib import Path

import bluepyopt

from bluepyemodel.emodel_pipeline.emodel import EModel
from bluepyemodel.evaluation.evaluation import get_evaluator_from_access_point
from bluepyemodel.tools.utils import get_checkpoint_path
from bluepyemodel.tools.utils import logger
from bluepyemodel.tools.utils import read_checkpoint

logger = logging.getLogger(__name__)


def setup_optimiser(
    evaluator, map_function, params, optimiser="IBEA", use_stagnation_criterion=True
):
    """Setup the bluepyopt optimiser.

    Args:
        evaluator (CellEvaluator): evaluator used to compute the scores.
        map_function (map): used to parallelize the evaluation of the
            individual in the population.
        params (dict): optimisation meta-parameters.
        optimiser (str): name of the optimiser, has to be "IBEA", "SO-CMA" or
            "MO-CMA".
        use_stagnation_criterion (bool): whether to use the stagnation
            stopping criterion on top of the maximum generation criterion for CMA

    Returns:
        DEAPOptimisation
    """
    if optimiser == "IBEA":
        return bluepyopt.deapext.optimisations.IBEADEAPOptimisation(
            evaluator=evaluator, map_function=map_function, **params
        )
    if optimiser == "SO-CMA":
        return bluepyopt.deapext.optimisationsCMA.DEAPOptimisationCMA(
            evaluator=evaluator,
            map_function=map_function,
            selector_name="single_objective",
            use_stagnation_criterion=use_stagnation_criterion,
            **params,
        )
    if optimiser == "MO-CMA":
        return bluepyopt.deapext.optimisationsCMA.DEAPOptimisationCMA(
            evaluator=evaluator,
            map_function=map_function,
            selector_name="multi_objective",
            use_stagnation_criterion=use_stagnation_criterion,
            **params,
        )
    raise ValueError(f"Unknown optimiser: {optimiser}")


def run_optimisation(
    optimiser,
    checkpoint_path,
    max_ngen,
    terminator=None,
    optimisation_checkpoint_period=None,
):
    """Run the optimisation.

    Args:
        optimiser (DEAPOptimisation): optimiser used for the run.
        checkpoint_path (str): path to which the checkpoint will be saved.
        max_ngen (int): maximum number of generation for which the
            evolutionary strategy will run.
        terminator (multiprocessing.Event): end optimisation when is set.
            Not taken into account if None.

    Returns:
        None
    """

    Path(checkpoint_path).parents[0].mkdir(parents=True, exist_ok=True)

    if os.path.isfile(checkpoint_path):
        logger.info(
            "Checkopint already exists."
            "Will continue optimisation from last generation in checkpoint"
        )
        continue_opt = True
    else:
        logger.info("No checkpoint found. Will start optimisation from scratch.")
        continue_opt = False

    logger.info("Running optimisation ...")
    pop, hof, log, history = optimiser.run(
        max_ngen=max_ngen,
        cp_filename=str(checkpoint_path),
        cp_period=optimisation_checkpoint_period,
        continue_cp=continue_opt,
        terminator=terminator,
    )
    logger.info("Running optimisation ... Done.")

    return pop, hof, log, history


def setup_and_run_optimisation(
    access_point,
    seed,
    mapper=None,
    terminator=None,
):
    cell_evaluator = get_evaluator_from_access_point(
        access_point=access_point, include_validation_protocols=False
    )

    use_stagnation_criterion = access_point.pipeline_settings.use_stagnation_criterion
    opt_params = access_point.pipeline_settings.optimisation_params
    if "centroids" in opt_params and isinstance(opt_params["centroids"][0], dict):
        opt_params["centroids"][0] = [
            opt_params["centroids"][0][name] for name in list(cell_evaluator.param_names)
        ]

    opt_params["seed"] = seed

    opt = setup_optimiser(
        cell_evaluator,
        mapper,
        params=opt_params,
        optimiser=access_point.pipeline_settings.optimiser,
        use_stagnation_criterion=use_stagnation_criterion,
    )

    checkpoint_path = get_checkpoint_path(access_point.emodel_metadata, seed)

    optimisation_checkpoint_period = access_point.pipeline_settings.optimisation_checkpoint_period
    run_optimisation(
        optimiser=opt,
        checkpoint_path=checkpoint_path,
        max_ngen=access_point.pipeline_settings.max_ngen,
        terminator=terminator,
        optimisation_checkpoint_period=optimisation_checkpoint_period,
    )


def store_best_model(
    access_point,
    seed=None,
    checkpoint_path=None,
):
    """Store the best model from an optimisation. Reads a checkpoint file generated
        by BluePyOpt and store the best individual of the hall of fame.

    Args:
        access_point (DataAccessPoint): data access point.
        checkpoint_path (str): path to the checkpoint file. If None, checkpoint_dir will be used.
    """

    cell_evaluator = get_evaluator_from_access_point(
        access_point=access_point,
        include_validation_protocols=False,
    )

    if checkpoint_path is None:
        if seed is None:
            raise TypeError("Please specify either the seed or the checkpoint_path")

        checkpoint_path = get_checkpoint_path(access_point.emodel_metadata, seed=seed)

    run, seed_from_checkpoint = read_checkpoint(checkpoint_path)

    best_model = run["halloffame"][0]
    feature_names = [obj.name for obj in cell_evaluator.fitness_calculator.objectives]

    if "param_names" in run:
        if run["param_names"] != list(cell_evaluator.param_names):
            raise ValueError(
                "The parameter names present in the checkpoint file are different from the "
                f"ones of the evaluator: {run['param_names']} versus "
                f"{list(cell_evaluator.param_names)}"
            )
        params = dict(zip(run["param_names"], best_model))
    else:
        params = dict(zip(list(cell_evaluator.param_names), best_model))

    scores = dict(zip(feature_names, best_model.fitness.values))

    emodel_seed = seed_from_checkpoint if seed is None else seed

    emodel = EModel(
        fitness=sum(list(scores.values())),
        parameter=params,
        score=scores,
        seed=emodel_seed,
        emodel_metadata=access_point.emodel_metadata,
    )

    access_point.store_emodel(emodel)



"""Auto-targets-related functions."""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

logger = logging.getLogger(__name__)


AUTO_TARGET_DICT = {
    "firing_pattern": {
        "protocols": [
            "Step",
            "FirePattern",
            "IDrest",
            "IDRest",
            "IDthresh",
            "IDThresh",
            "IDThres",
            "IDthres",
            "IV",
        ],
        "amplitudes": [200, 150, 250, 300],
        "efeatures": [
            "Spikecount",
            "depol_block_bool",
            "voltage_base",
            "voltage_after_stim",
            "mean_frequency",
            "time_to_first_spike",
            "time_to_last_spike",
            "inv_time_to_first_spike",
            "inv_first_ISI",
            "inv_second_ISI",
            "inv_third_ISI",
            "inv_last_ISI",
            "ISI_CV",
            "ISI_log_slope",
            "doublet_ISI",
            "AHP_depth",
            "AHP_time_from_peak",
            "strict_burst_number",
            "strict_burst_mean_freq",
            "number_initial_spikes",
            "irregularity_index",
            "adaptation_index",
        ],
        "min_recordings_per_amplitude": 1,
        "preferred_number_protocols": 2,
        "tolerance": 20.0,
    },
    "ap_waveform": {
        "protocols": [
            "APWaveform",
            "APwaveform",
            "Step",
            "FirePattern",
            "IDrest",
            "IDRest",
            "IDthresh",
            "IDThresh",
            "IDThres",
            "IDthres",
            "IV",
        ],
        "amplitudes": [300, 350, 250, 400, 200],
        "efeatures": [
            "AP_amplitude",
            "AP1_amp",
            "AP_duration_half_width",
            "AHP_depth",
        ],
        "min_recordings_per_amplitude": 1,
        "preferred_number_protocols": 1,
        "tolerance": 20.0,
    },
    "iv": {
        "protocols": ["IV", "Step"],
        "amplitudes": [0, -40, -100],
        "efeatures": [
            "voltage_base",
            "ohmic_input_resistance_vb_ssse",
        ],
        "min_recordings_per_amplitude": 1,
        "preferred_number_protocols": 2,
        "tolerance": 10.0,
    },
    "validation": {
        "protocols": ["IDhyperpol", "IDHyperpol"],
        "amplitudes": [150, 200],
        "efeatures": [
            "mean_frequency",
            "voltage_base",
            "depol_block_bool",
        ],
        "min_recordings_per_amplitude": 1,
        "preferred_number_protocols": 2,
        "tolerance": 20.0,
    },
}


def get_auto_target_from_presets(presets):
    """Returns auto-target given preset name.

    Args:
        presets (list of str): list of preset names
    """
    auto_targets = []
    for preset in presets:
        if preset in AUTO_TARGET_DICT:
            auto_targets.append(AUTO_TARGET_DICT[preset])
        else:
            logger.warning("Could not find auto-target preset %s", preset)
    return auto_targets



"""Luigi tasks for emodel optimisation."""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import glob
import logging
import multiprocessing
from pathlib import Path

import luigi

from bluepyemodel.access_point.access_point import OptimisationState
from bluepyemodel.efeatures_extraction.efeatures_extraction import extract_save_features_protocols
from bluepyemodel.efeatures_extraction.targets_configurator import TargetsConfigurator
from bluepyemodel.emodel_pipeline.plotting import optimisation
from bluepyemodel.emodel_pipeline.plotting import plot_models
from bluepyemodel.optimisation import get_checkpoint_path
from bluepyemodel.optimisation import store_best_model
from bluepyemodel.tasks.config import EmodelAPIConfig
from bluepyemodel.tasks.luigi_tools import IPyParallelTask
from bluepyemodel.tasks.luigi_tools import WorkflowTarget
from bluepyemodel.tasks.luigi_tools import WorkflowTask
from bluepyemodel.tasks.luigi_tools import WorkflowTaskRequiringMechanisms
from bluepyemodel.tasks.luigi_tools import WorkflowWrapperTask
from bluepyemodel.tools.mechanisms import compile_mechs_in_emodel_dir

# pylint: disable=W0235,W0621,W0404,W0611,W0703,E1128
logger = logging.getLogger(__name__)


def _reformat_ttype(ttype):
    """ """

    if isinstance(ttype, str):
        return ttype.replace("__", " ")

    return None


class CreateTargetsConfigurationTarget(WorkflowTarget):
    """Luigi target to check if BPEM targets configuraiton for extraction exists."""

    def __init__(self, emodel, etype, ttype, mtype, species, brain_region, iteration_tag):
        """Constructor."""
        super().__init__(
            emodel=emodel,
            etype=etype,
            ttype=ttype,
            mtype=mtype,
            species=species,
            brain_region=brain_region,
            iteration_tag=iteration_tag,
        )

    def exists(self):
        """Check if the targets configuration has been created."""
        return self.access_point.has_targets_configuration()


class CreateTargetsConfiguration(WorkflowTask):
    """Luigi wrapper for create_targets in emodel_pipeline.EModel_pipeline."""

    @WorkflowTask.check_mettypes
    def run(self):
        """ """

        configurator = TargetsConfigurator(access_point=self.access_point)
        configurator.create_and_save_configuration_from_access_point()

    def output(self):
        """ """
        return CreateTargetsConfigurationTarget(
            emodel=self.emodel,
            etype=self.etype,
            ttype=self.ttype,
            mtype=self.mtype,
            species=self.species,
            brain_region=self.brain_region,
            iteration_tag=self.iteration_tag,
        )


class EfeaturesProtocolsTarget(WorkflowTarget):
    """Target to check if efeatures and protocols are present in the database."""

    def __init__(self, emodel, etype, ttype, mtype, species, brain_region, iteration_tag):
        """Constructor."""
        super().__init__(
            emodel=emodel,
            etype=etype,
            ttype=ttype,
            mtype=mtype,
            species=species,
            brain_region=brain_region,
            iteration_tag=iteration_tag,
        )

    def exists(self):
        """Check if the features and protocols have been created."""

        return self.access_point.has_fitness_calculator_configuration()


class ExtractEFeatures(WorkflowTask):
    """Luigi wrapper for extract_efeatures in emodel_pipeline.EModel_pipeline."""

    @WorkflowTask.check_mettypes
    def run(self):
        """ """

        mapper = self.get_mapper()

        _ = extract_save_features_protocols(access_point=self.access_point, mapper=mapper)

    def output(self):
        """ """
        return EfeaturesProtocolsTarget(
            emodel=self.emodel,
            etype=self.etype,
            ttype=self.ttype,
            mtype=self.mtype,
            species=self.species,
            brain_region=self.brain_region,
            iteration_tag=self.iteration_tag,
        )


class CompileMechanisms(WorkflowTaskRequiringMechanisms):
    """Luigi wrapper for mechanisms.compile_mechs_in_emodel_dir"""

    def run(self):
        """ """

        mechanisms_directory = self.access_point.get_mechanisms_directory()
        compile_mechs_in_emodel_dir(mechanisms_directory)

    def output(self):
        """ """

        targets = []
        mechanisms_directory = self.access_point.get_mechanisms_directory()

        architectures = ["x86_64", "i686", "powerpc", "umac"]

        # Detect the first existing architecture
        arch_dir = "x86_64"  # default
        for arch in architectures:
            dir_to_check = mechanisms_directory.parents[0] / arch
            if dir_to_check.exists() and dir_to_check.is_dir():
                arch_dir = dir_to_check
                break

        for filepath in glob.glob(f"{str(mechanisms_directory)}/*.mod"):
            compile_path = mechanisms_directory.parents[0] / arch_dir / f"{Path(filepath).stem}.o"
            targets.append(luigi.LocalTarget(compile_path))

        return targets


class OptimisationTarget(WorkflowTarget):
    """Target to check if an optimisation is present in the database."""

    def __init__(
        self,
        emodel,
        etype,
        ttype,
        mtype,
        species,
        brain_region,
        iteration_tag,
        seed=1,
        continue_opt=False,
    ):
        """Constructor.

        Args:
           seed (int): seed used in the optimisation.
           continue_opt (bool): whether to continue optimisation or not
                when the optimisation is not complete.
        """
        super().__init__(
            emodel=emodel,
            etype=etype,
            ttype=ttype,
            mtype=mtype,
            species=species,
            brain_region=brain_region,
            iteration_tag=iteration_tag,
        )

        self.seed = seed
        self.continue_opt = continue_opt

    def exists(self):
        """Check if the model is completed."""
        state = self.access_point.optimisation_state(seed=self.seed, continue_opt=self.continue_opt)
        if state == OptimisationState.COMPLETED:
            return True

        return False


class Optimise(WorkflowTaskRequiringMechanisms, IPyParallelTask):
    """Luigi wrapper for emodel_pipeline.emodel_creation.optimise

    Parameters:
        seed (int): seed used in the optimisation.
        continue_unfinished_optimisation (bool): whether to continue optimisation or not
                when the optimisation is not complete.
        graceful_killer (multiprocessing.Event): event triggered when USR1 signal is received.
            Has to use multiprocessing event for communicating between processes
            when there is more than 1 luigi worker.
            When set, will gracefully exit main loop in Optimise (in deap algorithm).
    """

    # if default not set, crashes when parameters are read by luigi_tools.copy_params
    seed = luigi.IntParameter(default=42)
    continue_unfinished_optimisation = luigi.BoolParameter(default=False)
    graceful_killer = multiprocessing.Event()

    def requires(self):
        """ """
        compile_mechanisms = self.access_point.pipeline_settings.compile_mechanisms

        targets = [
            ExtractEFeatures(
                emodel=self.emodel,
                etype=self.etype,
                ttype=self.ttype,
                mtype=self.mtype,
                species=self.species,
                brain_region=self.brain_region,
                iteration_tag=self.iteration_tag,
            )
        ]

        if compile_mechanisms:
            targets.append(
                CompileMechanisms(
                    emodel=self.emodel,
                    etype=self.etype,
                    ttype=self.ttype,
                    mtype=self.mtype,
                    species=self.species,
                    brain_region=self.brain_region,
                    iteration_tag=self.iteration_tag,
                )
            )

        return targets

    @WorkflowTask.check_mettypes
    def run(self):
        """Prepare self.args, then call bbp-workflow's IPyParallelTask's run()."""
        attrs = [
            "backend",
            "emodel",
            "seed",
            "etype",
            "ttype",
            "mtype",
            "species",
            "brain_region",
            "iteration_tag",
        ]
        self.prepare_args_for_remote_script(attrs)

        super().run()

    def remote_script(self):
        """Catch arguments from parsing, and run optimisation."""
        # This function will be copied into a file, and then
        # arguments will be passed to it from command line
        # so arguments should be read using argparse.
        # All functions used should be imported.
        # Class methods and attributes cannot be used.

        # -- imports -- #
        import argparse
        import json

        from bluepyemodel import access_point
        from bluepyemodel.optimisation import setup_and_run_optimisation
        from bluepyemodel.tools.multiprocessing import get_mapper

        # -- parsing -- #
        parser = argparse.ArgumentParser()
        parser.add_argument("--backend", default=None, type=str)
        parser.add_argument("--api_from_config", default="local", type=str)
        parser.add_argument(
            "--api_args_from_config",
            default=", ".join(
                (
                    '{"emodel_dir": None',
                    '"recipes_path": None',
                    '"final_path": None',
                    '"legacy_dir_structure": None}',
                )
            ),
            type=json.loads,
        )
        parser.add_argument("--emodel", default=None, type=str)
        parser.add_argument("--etype", default=None, type=str)
        parser.add_argument("--ttype", default=None, type=str)
        parser.add_argument("--mtype", default=None, type=str)
        parser.add_argument("--species", default=None, type=str)
        parser.add_argument("--brain_region", default=None, type=str)
        parser.add_argument("--iteration_tag", default=None, type=str)
        parser.add_argument("--seed", default=1, type=int)
        parser.add_argument("--ipyparallel_profile", default=None, type=str)
        args = parser.parse_args()

        # -- run optimisation -- #
        mapper = get_mapper(args.backend, ipyparallel_profile=args.ipyparallel_profile)
        access_pt = access_point.get_access_point(
            access_point=args.api_from_config,
            emodel=args.emodel,
            etype=args.etype,
            ttype=args.ttype,
            mtype=args.mtype,
            species=args.species,
            brain_region=args.brain_region,
            iteration_tag=args.iteration_tag,
            **args.api_args_from_config,
        )
        setup_and_run_optimisation(access_pt, args.seed, mapper=mapper)

    def output(self):
        """ """
        return OptimisationTarget(
            emodel=self.emodel,
            etype=self.etype,
            ttype=self.ttype,
            mtype=self.mtype,
            species=self.species,
            brain_region=self.brain_region,
            iteration_tag=self.iteration_tag,
            seed=self.seed,
            continue_opt=self.continue_unfinished_optimisation,
        )


class BestModelTarget(WorkflowTarget):
    """Check if the best model from optimisation is present in the database."""

    def __init__(self, emodel, etype, ttype, mtype, species, brain_region, iteration_tag, seed=1):
        """Constructor.

        Args:
            seed (int): seed used in the optimisation.
        """
        super().__init__(
            emodel=emodel,
            etype=etype,
            ttype=ttype,
            mtype=mtype,
            species=species,
            brain_region=brain_region,
            iteration_tag=iteration_tag,
        )

        self.seed = seed

    def exists(self):
        """Check if the best model is stored."""
        return self.access_point.has_best_model(seed=self.seed)


class StoreBestModels(WorkflowTaskRequiringMechanisms):
    """Luigi wrapper for store_best_model.

    Parameters:
        seed (int): seed used in the optimisation.
    """

    seed = luigi.IntParameter(default=1)

    def requires(self):
        """ """
        batch_size = self.access_point.pipeline_settings.optimisation_batch_size

        to_run = []

        for seed in range(self.seed, self.seed + batch_size):
            to_run.append(
                Optimise(
                    emodel=self.emodel,
                    etype=self.etype,
                    ttype=self.ttype,
                    mtype=self.mtype,
                    species=self.species,
                    brain_region=self.brain_region,
                    iteration_tag=self.iteration_tag,
                    seed=seed,
                )
            )

            to_run.append(
                PlotOptimisation(
                    emodel=self.emodel,
                    etype=self.etype,
                    ttype=self.ttype,
                    mtype=self.mtype,
                    species=self.species,
                    brain_region=self.brain_region,
                    iteration_tag=self.iteration_tag,
                    seed=seed,
                )
            )

        to_run.append(
            ExtractEFeatures(
                emodel=self.emodel,
                etype=self.etype,
                ttype=self.ttype,
                mtype=self.mtype,
                species=self.species,
                brain_region=self.brain_region,
                iteration_tag=self.iteration_tag,
            )
        )

        return to_run

    @WorkflowTask.check_mettypes
    def run(self):
        """ """
        batch_size = self.access_point.pipeline_settings.optimisation_batch_size
        for seed in range(self.seed, self.seed + batch_size):
            # can have unfulfilled dependecies if slurm has send signal near time limit.
            if OptimisationTarget(
                emodel=self.emodel,
                etype=self.etype,
                ttype=self.ttype,
                mtype=self.mtype,
                species=self.species,
                brain_region=self.brain_region,
                iteration_tag=self.iteration_tag,
                seed=seed,
            ).exists():
                store_best_model(self.access_point, seed)

    def output(self):
        """ """
        batch_size = self.access_point.pipeline_settings.optimisation_batch_size
        targets = []
        for seed in range(self.seed, self.seed + batch_size):
            targets.append(
                BestModelTarget(
                    emodel=self.emodel,
                    etype=self.etype,
                    ttype=self.ttype,
                    mtype=self.mtype,
                    species=self.species,
                    brain_region=self.brain_region,
                    iteration_tag=self.iteration_tag,
                    seed=seed,
                )
            )
        return targets


class ValidationTarget(WorkflowTarget):
    """Check if validation has been performed on the model.

    Return True if Validation task has already been performed on the model,
        even if the model is not validated.
    """

    def __init__(self, emodel, etype, ttype, mtype, species, brain_region, iteration_tag, seed):
        """Constructor.

        Args:
            emodel (str): name of the emodel. Has to match the name of the emodel
                under which the configuration data are stored.
            ttype (str): name of the ttype.
            seed (int): seed used in the optimisation.
        """
        super().__init__(
            emodel=emodel,
            etype=etype,
            ttype=ttype,
            mtype=mtype,
            species=species,
            brain_region=brain_region,
            iteration_tag=iteration_tag,
        )

        self.seed = seed

    def exists(self):
        """Check if the model is completed for all given seeds."""
        batch_size = self.access_point.pipeline_settings.optimisation_batch_size
        checked_for_all_seeds = [
            self.access_point.is_checked_by_validation(seed=seed)
            for seed in range(self.seed, self.seed + batch_size)
        ]
        return all(checked_for_all_seeds)


class Validation(WorkflowTaskRequiringMechanisms, IPyParallelTask):
    """Luigi wrapper for validation.

    Parameters:
        seed (int): seed used in the optimisation.
        graceful_killer (multiprocessing.Event): event triggered when USR1 signal is received.
            Has to use multiprocessing event for communicating between processes
            when there is more than 1 luigi worker. Skip task if set.
    """

    seed = luigi.IntParameter(default=1)
    graceful_killer = multiprocessing.Event()

    def requires(self):
        """ """

        plot_optimisation = self.access_point.pipeline_settings.plot_optimisation
        compile_mechanisms = self.access_point.pipeline_settings.compile_mechanisms

        to_run = [
            StoreBestModels(
                emodel=self.emodel,
                etype=self.etype,
                ttype=self.ttype,
                mtype=self.mtype,
                species=self.species,
                brain_region=self.brain_region,
                iteration_tag=self.iteration_tag,
                seed=self.seed,
            )
        ]
        if compile_mechanisms:
            to_run.append(
                CompileMechanisms(
                    emodel=self.emodel,
                    etype=self.etype,
                    ttype=self.ttype,
                    mtype=self.mtype,
                    species=self.species,
                    brain_region=self.brain_region,
                    iteration_tag=self.iteration_tag,
                )
            )

        if plot_optimisation:
            to_run.append(
                PlotModels(
                    emodel=self.emodel,
                    etype=self.etype,
                    ttype=self.ttype,
                    mtype=self.mtype,
                    species=self.species,
                    brain_region=self.brain_region,
                    iteration_tag=self.iteration_tag,
                    seed=self.seed,
                )
            )

        to_run.append(
            ExtractEFeatures(
                emodel=self.emodel,
                etype=self.etype,
                ttype=self.ttype,
                mtype=self.mtype,
                species=self.species,
                brain_region=self.brain_region,
                iteration_tag=self.iteration_tag,
            )
        )

        return to_run

    @WorkflowTask.check_mettypes
    def run(self):
        """Prepare self.args, then call bbp-workflow's IPyParallelTask's run()."""
        attrs = [
            "backend",
            "emodel",
            "etype",
            "ttype",
            "mtype",
            "species",
            "brain_region",
            "iteration_tag",
        ]

        self.prepare_args_for_remote_script(attrs)

        super().run()

    def remote_script(self):
        """Catch arguments from parsing, and run validation."""
        # -- imports -- #
        import argparse
        import json

        from bluepyemodel import access_point
        from bluepyemodel.tools.multiprocessing import get_mapper
        from bluepyemodel.validation.validation import validate

        # -- parsing -- #
        parser = argparse.ArgumentParser()
        parser.add_argument("--backend", default=None, type=str)
        parser.add_argument("--api_from_config", default="local", type=str)
        parser.add_argument(
            "--api_args_from_config",
            default=", ".join(
                (
                    '{"emodel_dir": None',
                    '"recipes_path": None',
                    '"final_path": None',
                    '"legacy_dir_structure": None}',
                )
            ),
            type=json.loads,
        )
        parser.add_argument("--emodel", default=None, type=str)
        parser.add_argument("--etype", default=None, type=str)
        parser.add_argument("--ttype", default=None, type=str)
        parser.add_argument("--mtype", default=None, type=str)
        parser.add_argument("--species", default=None, type=str)
        parser.add_argument("--brain_region", default=None, type=str)
        parser.add_argument("--iteration_tag", default=None, type=str)
        parser.add_argument("--ipyparallel_profile", default=None, type=str)

        args = parser.parse_args()

        # -- run validation -- #
        mapper = get_mapper(args.backend, ipyparallel_profile=args.ipyparallel_profile)
        access_pt = access_point.get_access_point(
            access_point=args.api_from_config,
            emodel=args.emodel,
            etype=args.etype,
            ttype=args.ttype,
            mtype=args.mtype,
            species=args.species,
            brain_region=args.brain_region,
            iteration_tag=args.iteration_tag,
            **args.api_args_from_config,
        )

        validate(access_pt, mapper)

    def output(self):
        """ """
        return ValidationTarget(
            emodel=self.emodel,
            etype=self.etype,
            ttype=self.ttype,
            mtype=self.mtype,
            species=self.species,
            brain_region=self.brain_region,
            iteration_tag=self.iteration_tag,
            seed=self.seed,
        )


class StoreHocTarget(WorkflowTarget):
    """Check if the hoc files have been stored."""

    def __init__(self, emodel, etype, ttype, mtype, species, brain_region, iteration_tag, seed):
        """Constructor.

        Args:
            emodel (str): name of the emodel. Has to match the name of the emodel
                under which the configuration data are stored.
            ttype (str): name of the ttype.
            seed (int): seed used in the optimisation.
        """
        super().__init__(
            emodel=emodel,
            etype=etype,
            ttype=ttype,
            mtype=mtype,
            species=species,
            brain_region=brain_region,
            iteration_tag=iteration_tag,
        )

        self.seed = seed

    def exists(self):
        """Check if the hoc is stored for all given seeds."""
        batch_size = self.access_point.pipeline_settings.optimisation_batch_size
        checked_for_all_seeds = [
            self.access_point.sonata_exists(seed=seed)
            for seed in range(self.seed, self.seed + batch_size)
        ]
        return all(checked_for_all_seeds)


class ExportHoc(WorkflowTaskRequiringMechanisms, IPyParallelTask):
    """Luigi wrapper for export_emodels_sonata.

    Parameters:
        seed (int): seed used in the optimisation.
        graceful_killer (multiprocessing.Event): event triggered when USR1 signal is received.
            Has to use multiprocessing event for communicating between processes
            when there is more than 1 luigi worker. Skip task if set.
    """

    seed = luigi.IntParameter(default=1)
    graceful_killer = multiprocessing.Event()

    def __init__(self, *args, **kwargs):
        """ """
        super().__init__(*args, **kwargs)

        # set self.batch_size here to easily handle it in the argparse argument passing
        self.batch_size = self.access_point.pipeline_settings.optimisation_batch_size

    def requires(self):
        """ """

        compile_mechanisms = self.access_point.pipeline_settings.compile_mechanisms

        to_run = [
            StoreBestModels(
                emodel=self.emodel,
                etype=self.etype,
                ttype=self.ttype,
                mtype=self.mtype,
                species=self.species,
                brain_region=self.brain_region,
                iteration_tag=self.iteration_tag,
                seed=self.seed,
            )
        ]
        if compile_mechanisms:
            to_run.append(
                CompileMechanisms(
                    emodel=self.emodel,
                    etype=self.etype,
                    ttype=self.ttype,
                    mtype=self.mtype,
                    species=self.species,
                    brain_region=self.brain_region,
                    iteration_tag=self.iteration_tag,
                )
            )

        return to_run

    @WorkflowTask.check_mettypes
    def run(self):
        """Prepare self.args, then call bbp-workflow's IPyParallelTask's run()."""
        attrs = [
            "backend",
            "emodel",
            "etype",
            "ttype",
            "mtype",
            "species",
            "brain_region",
            "iteration_tag",
            "seed",
            "batch_size",
        ]

        self.prepare_args_for_remote_script(attrs)

        super().run()

    def remote_script(self):
        """Catch arguments from parsing, and run validation."""
        # -- imports -- #
        import argparse
        import json

        from bluepyemodel import access_point
        from bluepyemodel.export_emodel.export_emodel import export_emodels_sonata
        from bluepyemodel.tools.multiprocessing import get_mapper

        # -- parsing -- #
        parser = argparse.ArgumentParser()
        parser.add_argument("--backend", default=None, type=str)
        parser.add_argument("--api_from_config", default="local", type=str)
        parser.add_argument(
            "--api_args_from_config",
            default=", ".join(
                (
                    '{"emodel_dir": None',
                    '"recipes_path": None',
                    '"final_path": None',
                    '"legacy_dir_structure": None}',
                )
            ),
            type=json.loads,
        )
        parser.add_argument("--emodel", default=None, type=str)
        parser.add_argument("--etype", default=None, type=str)
        parser.add_argument("--ttype", default=None, type=str)
        parser.add_argument("--mtype", default=None, type=str)
        parser.add_argument("--species", default=None, type=str)
        parser.add_argument("--brain_region", default=None, type=str)
        parser.add_argument("--iteration_tag", default=None, type=str)
        parser.add_argument("--ipyparallel_profile", default=None, type=str)
        parser.add_argument("--seed", default=None, type=int)
        parser.add_argument("--batch_size", default=None, type=int)

        args = parser.parse_args()

        # -- run validation -- #
        mapper = get_mapper(args.backend, ipyparallel_profile=args.ipyparallel_profile)
        access_pt = access_point.get_access_point(
            access_point=args.api_from_config,
            emodel=args.emodel,
            etype=args.etype,
            ttype=args.ttype,
            mtype=args.mtype,
            species=args.species,
            brain_region=args.brain_region,
            iteration_tag=args.iteration_tag,
            **args.api_args_from_config,
        )

        export_emodels_sonata(
            access_pt,
            only_validated=False,
            only_best=False,
            seeds=list(range(args.seed, args.seed + args.batch_size)),
            map_function=mapper,
        )
        if args.api_from_config == "nexus":
            access_pt.store_emodels_sonata(
                only_validated=False,
                only_best=False,
                seeds=list(range(args.seed, args.seed + args.batch_size)),
                map_function=mapper,
            )

    def output(self):
        """ """
        return StoreHocTarget(
            emodel=self.emodel,
            etype=self.etype,
            ttype=self.ttype,
            mtype=self.mtype,
            species=self.species,
            brain_region=self.brain_region,
            iteration_tag=self.iteration_tag,
            seed=self.seed,
        )


class EModelCreationTarget(WorkflowTarget):
    """Check if the the model is validated for any seed."""

    def __init__(self, emodel, etype, ttype, mtype, species, brain_region, iteration_tag):
        """Constructor.

        Args:
            emodel (str): name of the emodel. Has to match the name of the emodel
                under which the configuration data are stored.
            ttype (str): name of the ttype.
        """
        super().__init__(
            emodel=emodel,
            etype=etype,
            ttype=ttype,
            mtype=mtype,
            species=species,
            brain_region=brain_region,
            iteration_tag=iteration_tag,
        )

    def update_emodel_workflow(self, state="done"):
        """Update emodel workflow state on nexus"""
        # get / create
        emodel_workflow, _ = self.access_point.get_emodel_workflow()
        if emodel_workflow is None:
            emodel_workflow = self.access_point.create_emodel_workflow()

        # store / update
        # this function might be called several times by luigi, so
        # only update the resource when necessary
        if emodel_workflow.state != state:
            emodel_workflow.state = state
            self.access_point.store_or_update_emodel_workflow(emodel_workflow)

    def exists(self):
        """Check if the model is completed."""
        exist = self.access_point.is_validated()
        if exist:
            # only for nexus access_point
            self.update_emodel_workflow(state="done")

        return exist


class EModelCreation(WorkflowTask):
    """Main Workflow Task. Creates an emodel.

    Parameters:
        emodel (str): name of the emodel.
        ttype (str): name of the t-type.
        species (str): name of the species.
        brain_region (str): name of the brain region.
        seed (int): seed used in the optimisation.
        graceful_killer (multiprocessing.Event): event triggered when USR1 signal is received.
            Has to use multiprocessing event for communicating between processes
            when there is more than 1 luigi worker. Exit loop if set.
    """

    # pylint: disable=no-self-argument, not-callable

    seed = luigi.IntParameter(default=1)
    graceful_killer = multiprocessing.Event()

    def check_mettypes(func):
        """Decorator to check mtype, etype and ttype presence on nexus"""

        def inner(self):
            """Inner decorator function"""
            if EmodelAPIConfig().api == "nexus":
                self.access_point.check_mettypes()
            # do this instead of just func(self) because of the yield in EModelCreation
            yield from func(self)

        return inner

    def check_and_update_emodel_workflow(self, state="running"):
        """Get or create emodel workflow, check its configuration, and update its state on nexus"""
        # get / create
        emodel_workflow, _ = self.access_point.get_emodel_workflow()
        if emodel_workflow is None:
            emodel_workflow = self.access_point.create_emodel_workflow()

        # check
        has_configuration = self.access_point.check_emodel_workflow_configurations(emodel_workflow)
        if not has_configuration:
            raise ValueError(
                "There are configuration files missing on nexus for the workflow."
                "Please check that you have registered on nexus the following resources:\n"
                "ExtractionTargetsConfiguration\n"
                "EModelPipelineSettings\n"
                "EModelConfiguration"
            )

        # store / update
        emodel_workflow.state = state
        self.access_point.store_or_update_emodel_workflow(emodel_workflow)

    @check_mettypes
    def run(self):
        """Optimise e-models by batches of batch_size until one is validated."""

        seed = self.seed

        batch_size = self.access_point.pipeline_settings.optimisation_batch_size
        max_n_batch = self.access_point.pipeline_settings.max_n_batch

        # only for nexus access_point
        self.check_and_update_emodel_workflow(state="running")

        while not self.output().exists() and not self.graceful_killer.is_set():
            # limit the number of batch
            if seed > self.seed + max_n_batch * batch_size:
                break

            yield (
                Validation(
                    emodel=self.emodel,
                    etype=self.etype,
                    ttype=self.ttype,
                    mtype=self.mtype,
                    species=self.species,
                    brain_region=self.brain_region,
                    iteration_tag=self.iteration_tag,
                    seed=seed,
                ),
                ExportHoc(
                    emodel=self.emodel,
                    etype=self.etype,
                    ttype=self.ttype,
                    mtype=self.mtype,
                    species=self.species,
                    brain_region=self.brain_region,
                    iteration_tag=self.iteration_tag,
                    seed=seed,
                ),
            )
            seed += batch_size

        assert self.output().exists()

    def output(self):
        """ """
        return EModelCreationTarget(
            emodel=self.emodel,
            etype=self.etype,
            ttype=self.ttype,
            mtype=self.mtype,
            species=self.species,
            brain_region=self.brain_region,
            iteration_tag=self.iteration_tag,
        )


class SeveralEModelCreation(luigi.WrapperTask):
    """Wrapper around EModelCreation. Creates several emodels.

    Parameters:
        emodels (str): name of the emodels, separated by ",".
        etypes (str): None, etype or list of etypes of the emodels, separated by ",". Optional.
        ttypes (str): None, ttype or list of ttypes of the emodels, separated by ",". Optional.
        mtypes (str): None, mtype or list of mtypes of the emodels, separated by ",". Optional.
        species (str): name of the species. Optional.
        brain_region (str): name of the brain region. Optional.
        iteration_tag (str): iteration tag or githash to identify the run. Optional.
    """

    emodels = luigi.Parameter(default=[])
    etypes = luigi.Parameter(default=None)
    ttypes = luigi.Parameter(default=None)
    mtypes = luigi.Parameter(default=None)
    species = luigi.Parameter(default=None)
    brain_region = luigi.Parameter(default=None)
    iteration_tag = luigi.Parameter(default=None)

    @staticmethod
    def _parse_emodel_metadata(metadata, n_models):
        """Parse the string "etype1,etype2,..." into ["etype1", "etype2", ...]"""
        # pylint: disable=no-member, unsupported-membership-test
        if metadata is None or "," not in metadata:
            return [metadata] * n_models
        return metadata.split(",")

    def requires(self):
        # pylint: disable=no-member
        emodels = self.emodels.split(",")
        etypes = self._parse_emodel_metadata(self.etypes, len(emodels))
        ttypes = self._parse_emodel_metadata(self.ttypes, len(emodels))
        mtypes = self._parse_emodel_metadata(self.mtypes, len(emodels))

        to_run = []
        for emodel, etype, ttype, mtype in zip(emodels, etypes, ttypes, mtypes):
            to_run.append(
                EModelCreation(
                    emodel=emodel,
                    etype=etype,
                    ttype=ttype,
                    mtype=mtype,
                    species=self.species,
                    brain_region=self.brain_region,
                    iteration_tag=self.iteration_tag,
                )
            )

        return to_run


class EModelCreationWrapper(WorkflowWrapperTask):
    """Luigi wrapper for launching EModel Creation pipeline.

    For now, launches only one EModel Creation pipeline. Could be changed later.

    Parameters:
        emodel (str): name of the emodel. Has to match the name of the emodel
            under which the configuration data are stored.
        ttype (str): name of the t-type.
        species (str): name of the species.
        brain_region (str): name of the brain region.
    """

    species = luigi.Parameter(default=None)
    brain_region = luigi.Parameter(default=None)

    def requires(self):
        """ """
        to_run = [
            EModelCreation(
                emodel=self.emodel,
                etype=self.etype,
                ttype=self.ttype,
                mtype=self.mtype,
                species=self.species,
                brain_region=self.brain_region,
                iteration_tag=self.iteration_tag,
            )
        ]

        plot_optimisation = self.access_point.pipeline_settings.plot_optimisation

        if plot_optimisation:
            to_run.append(
                PlotValidatedDistributions(
                    emodel=self.emodel,
                    etype=self.etype,
                    ttype=self.ttype,
                    mtype=self.mtype,
                    species=self.species,
                    brain_region=self.brain_region,
                    iteration_tag=self.iteration_tag,
                )
            )

        return to_run


class OptimiseWrapper(WorkflowWrapperTask):
    """Luigi wrapper for launching multiple seeds to optimise.

    Parameters:
        emodel (str): name of the emodel. Has to match the name of the emodel
            under which the configuration data are stored.
        ttype (str): name of the t-type.
        species (str): name of the species.
        brain_region (str): name of the brain region.
        seed (int): seed used in the optimisation.
        batch_size (int): number of seeds to optimise at the same time before each validation.
    """

    seed = luigi.IntParameter(default=42)

    def requires(self):
        """ """
        to_run = []

        batch_size = self.access_point.pipeline_settings.optimisation_batch_size

        for seed in range(self.seed, self.seed + batch_size):
            to_run.append(
                Optimise(
                    emodel=self.emodel,
                    etype=self.etype,
                    ttype=self.ttype,
                    mtype=self.mtype,
                    species=self.species,
                    brain_region=self.brain_region,
                    iteration_tag=self.iteration_tag,
                    seed=seed,
                )
            )
        return to_run


class PlotOptimisation(WorkflowTask):
    """Luigi wrapper for plotting the optimisation outputs.

    Parameters:
        species (str): name of the species.
        brain_region (str): name of the brain region.
        seed (int): seed used in the optimisation.
        checkpoint_dir (str): path to the repo where files used as a checkpoint by BluePyOpt are.
    """

    seed = luigi.IntParameter(default=42)

    def requires(self):
        """ """
        return Optimise(
            emodel=self.emodel,
            etype=self.etype,
            ttype=self.ttype,
            mtype=self.mtype,
            species=self.species,
            brain_region=self.brain_region,
            iteration_tag=self.iteration_tag,
            seed=self.seed,
        )

    def run(self):
        """ """

        checkpoint_path = get_checkpoint_path(self.access_point.emodel_metadata, seed=self.seed)

        optimisation(
            optimiser=self.access_point.pipeline_settings.optimiser,
            emodel=self.emodel,
            iteration=self.iteration_tag,
            seed=self.seed,
            checkpoint_path=checkpoint_path,
            figures_dir=Path("./figures") / self.emodel / "optimisation",
        )

    def output(self):
        """ """
        checkpoint_path = get_checkpoint_path(self.access_point.emodel_metadata, seed=self.seed)

        fname = f"{Path(checkpoint_path).stem}.pdf"
        return luigi.LocalTarget(Path("./figures") / self.emodel / "optimisation" / fname)


class PlotModels(WorkflowTaskRequiringMechanisms):
    """Luigi wrapper for plotting the optimisation outputs.

    Parameters:
        species (str): name of the species.
        brain_region (str): name of the brain region.
        seed (int): seed used in the optimisation.
    """

    species = luigi.Parameter(default="")
    brain_region = luigi.Parameter(default="")
    seed = luigi.IntParameter(default=42)

    def requires(self):
        """ """
        requires = [
            ExtractEFeatures(
                emodel=self.emodel,
                etype=self.etype,
                ttype=self.ttype,
                mtype=self.mtype,
                species=self.species,
                brain_region=self.brain_region,
                iteration_tag=self.iteration_tag,
            ),
            StoreBestModels(
                emodel=self.emodel,
                etype=self.etype,
                ttype=self.ttype,
                mtype=self.mtype,
                species=self.species,
                brain_region=self.brain_region,
                iteration_tag=self.iteration_tag,
                seed=self.seed,
            ),
            CompileMechanisms(
                emodel=self.emodel,
                etype=self.etype,
                ttype=self.ttype,
                mtype=self.mtype,
                species=self.species,
                brain_region=self.brain_region,
                iteration_tag=self.iteration_tag,
            ),
        ]

        return requires

    def run(self):
        """ """

        plot_optimisation = self.access_point.pipeline_settings.plot_optimisation
        plot_currentscape = self.access_point.pipeline_settings.plot_currentscape
        plot_bAP_EPSP = self.access_point.pipeline_settings.plot_bAP_EPSP
        batch_size = self.access_point.pipeline_settings.optimisation_batch_size

        mapper = self.get_mapper()
        plot_models(
            access_point=self.access_point,
            mapper=mapper,
            seeds=range(self.seed, self.seed + batch_size),
            figures_dir=Path("./figures") / self.emodel,
            plot_distributions=plot_optimisation,
            plot_traces=plot_optimisation,
            plot_scores=plot_optimisation,
            plot_thumbnail=plot_optimisation,
            plot_dendritic_ISI_CV=plot_optimisation,
            plot_dendritic_rheobase=plot_optimisation,
            plot_currentscape=plot_currentscape,
            plot_bAP_EPSP=plot_bAP_EPSP,
        )

    def output(self):
        """ """

        batch_size = self.access_point.pipeline_settings.optimisation_batch_size
        plot_optimisation = self.access_point.pipeline_settings.plot_optimisation

        outputs = []
        if plot_optimisation:
            # distribution
            fname = self.access_point.emodel_metadata.as_string()
            fname += "__parameters_distribution.pdf"
            fpath = Path("./figures") / self.emodel / "distributions" / "all" / fname
            outputs.append(luigi.LocalTarget(fpath))

            # scores
            for seed in range(self.seed, self.seed + batch_size):
                fname = self.access_point.emodel_metadata.as_string(seed)
                fname += "__scores.pdf"
                fpath = Path("./figures") / self.emodel / "scores" / "all" / fname
                outputs.append(luigi.LocalTarget(fpath))

            # traces
            for seed in range(self.seed, self.seed + batch_size):
                fname = self.access_point.emodel_metadata.as_string(seed)
                fname += "__traces.pdf"
                fpath = Path("./figures") / self.emodel / "traces" / "all" / fname
                outputs.append(luigi.LocalTarget(fpath))

        return outputs


class PlotValidatedDistributions(WorkflowTaskRequiringMechanisms):
    """Luigi wrapper for plotting the optimisation outputs.

    Parameters:
        species (str): name of the species.
        brain_region (str): name of the brain region.
    """

    def requires(self):
        """ """

        return [
            EModelCreation(
                emodel=self.emodel,
                etype=self.etype,
                ttype=self.ttype,
                mtype=self.mtype,
                species=self.species,
                brain_region=self.brain_region,
                iteration_tag=self.iteration_tag,
            ),
            CompileMechanisms(
                emodel=self.emodel,
                etype=self.etype,
                ttype=self.ttype,
                mtype=self.mtype,
                species=self.species,
                brain_region=self.brain_region,
                iteration_tag=self.iteration_tag,
            ),
        ]

    def run(self):
        """ """
        plot_models(
            access_point=self.access_point,
            mapper=self.get_mapper(),
            figures_dir=Path("./figures") / self.emodel,
            plot_distributions=True,
            plot_traces=False,
            plot_scores=False,
            plot_thumbnail=False,
            plot_dendritic_ISI_CV=False,
            plot_dendritic_rheobase=False,
            plot_bAP_EPSP=False,
            only_validated=True,
        )

    def output(self):
        """ """

        fname = f"{self.emodel}_parameters_distribution.pdf"
        fpath = Path("./figures") / self.emodel / "distributions" / "validated" / fname

        return luigi.LocalTarget(fpath)



"""IDrest stimulus class"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

import numpy

from bluepyemodel.ecode.stimulus import BPEM_stimulus

logger = logging.getLogger(__name__)


class IDrest(BPEM_stimulus):
    """IDrest current stimulus

    .. code-block:: none

              holdi               holdi+amp                holdi
                :                     :                      :
                :                     :                      :
                :           ______________________           :
                :          |                      |          :
                :          |                      |          :
                :          |                      |          :
                :          |                      |          :
        |__________________|                      |______________________
        ^                  ^                      ^                      ^
        :                  :                      :                      :
        :                  :                      :                      :
        t=0                delay                  delay+duration         totduration
    """

    name = "IDrest"

    def __init__(self, location, **kwargs):
        """Constructor
        Args:
            location(Location): location of stimulus
        """

        self.amp = kwargs.get("amp", None)
        self.amp_rel = kwargs.get("thresh_perc", 200.0)

        self.holding_current = kwargs.get("holding_current", None)
        self.threshold_current = None

        if self.amp is None and self.amp_rel is None:
            raise TypeError(f"In stimulus {self.name}, amp and thresh_perc cannot be both None.")

        self.delay = kwargs.get("delay", 250.0)
        self.duration = kwargs.get("duration", 1350.0)
        self.total_duration = kwargs.get("totduration", 1850.0)

        super().__init__(
            location=location,
        )

    @property
    def stim_start(self):
        return self.delay

    @property
    def stim_end(self):
        return self.delay + self.duration

    @property
    def amplitude(self):
        if self.amp_rel is None or self.threshold_current is None:
            return self.amp
        return self.threshold_current * (float(self.amp_rel) / 100.0)

    def instantiate(self, sim=None, icell=None):
        """Run stimulus"""

        holding_current = self.holding_current if self.holding_current is not None else 0

        icomp = self.location.instantiate(sim=sim, icell=icell)

        self.iclamp = sim.neuron.h.IClamp(icomp.x, sec=icomp.sec)
        self.iclamp.dur = self.total_duration

        self.current_vec = sim.neuron.h.Vector()
        self.time_vec = sim.neuron.h.Vector()

        self.time_vec.append(0.0)
        self.current_vec.append(holding_current)

        self.time_vec.append(self.delay)
        self.current_vec.append(holding_current)

        self.time_vec.append(self.delay)
        self.current_vec.append(holding_current + self.amplitude)

        self.time_vec.append(self.stim_end)
        self.current_vec.append(holding_current + self.amplitude)

        self.time_vec.append(self.stim_end)
        self.current_vec.append(holding_current)

        self.time_vec.append(self.total_duration)
        self.current_vec.append(holding_current)

        self.iclamp.delay = 0
        self.current_vec.play(
            self.iclamp._ref_amp,  # pylint:disable=W0212
            self.time_vec,
            1,
            sec=icomp.sec,
        )

    def generate(self, dt=0.1):
        """Return current time series"""
        holding_current = self.holding_current if self.holding_current is not None else 0

        t = numpy.arange(0.0, self.total_duration, dt)
        current = numpy.full(t.shape, holding_current, dtype="float64")

        ton_idx = int(self.stim_start / dt)
        toff_idx = int(self.stim_end / dt)

        current[ton_idx:toff_idx] += self.amplitude

        return t, current



"""Export the emodels in the SONATA format"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import os
import pathlib
import shutil

import h5py

from bluepyemodel.evaluation.evaluation import compute_responses
from bluepyemodel.evaluation.evaluation import get_evaluator_from_access_point
from bluepyemodel.export_emodel.utils import get_hoc_file_path
from bluepyemodel.export_emodel.utils import get_output_path
from bluepyemodel.export_emodel.utils import select_emodels

logger = logging.getLogger(__name__)


def _write_node_file(emodel, model_template_path, node_file_path, morphology_path=None):
    """Creates a nodes.h5 file in the SONATA format. It contains the information
    needed to run the model. See https://bbpteam.epfl.ch/documentation/projects/
    circuit-documentation/latest/sonata_tech.html"""

    with h5py.File(node_file_path, "w") as f:
        population_name = f"{emodel.emodel_metadata.brain_region}_neurons"
        population = f.create_group(f"/nodes/{population_name}/0")

        population.create_dataset("x", (1,), dtype="f")
        population.create_dataset("y", (1,), dtype="f")
        population.create_dataset("z", (1,), dtype="f")

        population.create_dataset("orientation_w", (1,), dtype="f")
        population.create_dataset("orientation_x", (1,), dtype="f")
        population.create_dataset("orientation_y", (1,), dtype="f")
        population.create_dataset("orientation_z", (1,), dtype="f")

        population["model_template"] = str(model_template_path)
        population["model_type"] = "biophysical"

        if morphology_path:
            population["morphology"] = pathlib.Path(morphology_path).stem

        for k in ["etype", "mtype", "synapse_class"]:
            v = getattr(emodel.emodel_metadata, k)
            if v is not None:
                population[k] = v

        if emodel.emodel_metadata.brain_region is not None:
            population["region"] = emodel.emodel_metadata.brain_region

        threshold_keys = ("bpo_holding_current", "bpo_threshold_current")
        if all(key in emodel.responses for key in threshold_keys):
            dynamics_params = population.create_group("dynamics_params")
            dynamics_params["holding_current"] = emodel.responses["bpo_holding_current"]
            dynamics_params["threshold_current"] = emodel.responses["bpo_threshold_current"]


def _write_hoc_file(
    cell_model,
    emodel,
    hoc_file_path,
    template="cell_template_neurodamus_sbo.jinja2",
    new_emodel_name=None,
):
    """Creates a hoc file containing the emodel and its morphology.
    WARNING: this assumes that any morphology modifier has been informed as both
    a python method and a hoc method"""

    template_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "templates"))

    if new_emodel_name is not None:
        cell_model.name = new_emodel_name

    hoc_content = cell_model.create_hoc(
        param_values=emodel.parameters,
        template=template,
        template_dir=template_dir,
    )

    with open(hoc_file_path, "w") as f:
        f.writelines(hoc_content)


def _export_model_sonata(cell_model, emodel, output_dir=None, new_emodel_name=None):
    """Creates the directory and files required for an emodel to be used in circuit building"""

    if not emodel.passed_validation:
        logger.warning("Exporting a model that did not pass validation.")

    if new_emodel_name is not None:
        emodel.emodel_metadata.emodel = new_emodel_name

    output_path = get_output_path(emodel, output_dir, output_base_dir="export_emodels_sonata")
    hoc_file_path = get_hoc_file_path(output_path)
    node_file_path = str(output_path / "nodes.h5")
    morphology_path = str(output_path / pathlib.Path(cell_model.morphology.morphology_path).name)

    # Copy the morphology
    shutil.copyfile(cell_model.morphology.morphology_path, morphology_path)

    # Exports the BluePyOpt cell model as a hoc file
    _write_hoc_file(
        cell_model,
        emodel,
        hoc_file_path,
        template="cell_template_neurodamus_sbo.jinja2",
        new_emodel_name=new_emodel_name,
    )

    # Create the SONATA node file
    _write_node_file(
        emodel,
        model_template_path=hoc_file_path,
        node_file_path=node_file_path,
        morphology_path=morphology_path,
    )


def export_emodels_sonata(
    access_point,
    only_validated=False,
    only_best=True,
    seeds=None,
    map_function=map,
    new_emodel_name=None,
    new_metadata=None,
):
    """Export a set of emodels to a set of folder named after them. Each folder will
    contain a sonata nodes.h5 file, the morphology of the model and a hoc version of the model.
    """

    cell_evaluator = get_evaluator_from_access_point(
        access_point, include_validation_protocols=True
    )

    emodels = compute_responses(
        access_point,
        cell_evaluator,
        map_function,
        seeds=seeds,
        preselect_for_validation=False,
        store_responses=False,
    )

    emodels = select_emodels(
        access_point.emodel_metadata.emodel,
        emodels,
        only_validated=only_validated,
        only_best=only_best,
        seeds=seeds,
    )
    if not emodels:
        logger.warning(
            "No emodels were selected in export_emodels_sonata. Stopping sonata export here."
        )
        return

    cell_model = cell_evaluator.cell_model

    for mo in emodels:
        if new_metadata:
            mo.emodel_metadata = new_metadata
        if not cell_model.morphology.morph_modifiers:  # Turn [] into None
            cell_model.morphology.morph_modifiers = None
        _export_model_sonata(cell_model, mo, output_dir=None, new_emodel_name=new_emodel_name)


def _export_emodel_hoc(cell_model, mo, output_dir=None, new_emodel_name=None):
    if not mo.passed_validation:
        logger.warning("Exporting a model that did not pass validation.")

    if new_emodel_name is not None:
        mo.emodel_metadata.emodel = new_emodel_name

    output_path = get_output_path(mo, output_dir, output_base_dir="export_emodels_hoc")
    hoc_file_path = get_hoc_file_path(output_path)
    morphology_path = str(output_path / pathlib.Path(cell_model.morphology.morphology_path).name)

    # Copy the morphology
    shutil.copyfile(cell_model.morphology.morphology_path, morphology_path)

    # Exports the BluePyOpt cell model as a hoc file
    _write_hoc_file(
        cell_model,
        mo,
        hoc_file_path,
        template="cell_template.jinja2",
        new_emodel_name=new_emodel_name,
    )


def export_emodels_hoc(
    access_point,
    only_validated=False,
    only_best=True,
    seeds=None,
    new_emodel_name=None,
    new_metadata=None,
):
    """Export a set of emodels to a set of folder named after them. Each folder will contain a hoc
    version of the model."""

    cell_evaluator = get_evaluator_from_access_point(
        access_point, include_validation_protocols=True
    )

    emodels = access_point.get_emodels()

    emodels = select_emodels(
        access_point.emodel_metadata.emodel,
        emodels,
        only_validated=only_validated,
        only_best=only_best,
        seeds=seeds,
        iteration=access_point.emodel_metadata.iteration,
    )
    if not emodels:
        logger.warning("No emodels were selected in export_emodels_hoc. Stopping hoc export here.")
        return

    cell_model = cell_evaluator.cell_model

    for mo in emodels:
        if new_metadata:
            mo.emodel_metadata = new_metadata
        if not cell_model.morphology.morph_modifiers:  # Turn [] into None
            cell_model.morphology.morph_modifiers = None
        _export_emodel_hoc(cell_model, mo, output_dir=None, new_emodel_name=new_emodel_name)



"""Cell model creation."""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import collections
import importlib
import logging

from bluepyopt import ephys
from bluepyopt.ephys.morphologies import NrnFileMorphology
from bluepyopt.ephys.parameterscalers import NrnSegmentLinearScaler
from bluepyopt.ephys.parameterscalers import NrnSegmentSomaDistanceScaler
from bluepyopt.ephys.parameterscalers import NrnSegmentSomaDistanceStepScaler

from bluepyemodel.evaluation import modifiers
from bluepyemodel.evaluation.modifiers import replace_axon_hoc
from bluepyemodel.evaluation.modifiers import replace_axon_with_taper
from bluepyemodel.model.morphology_utils import get_hotspot_location

logger = logging.getLogger(__name__)


def multi_locations(section_name, additional_multiloc_map):
    """Define a list of locations from a section names.

    Args:
        section_name (str): name of section

    Returns:
        list: list of NrnSeclistLocation

    """

    multiloc_map = {
        "alldend": ["apical", "basal"],
        "somadend": ["apical", "basal", "somatic"],
        "allnoaxon": ["apical", "basal", "somatic"],
        "somaxon": ["axonal", "somatic"],
        "allact": ["apical", "basal", "somatic", "axonal"],
    }

    if additional_multiloc_map is not None:
        multiloc_map.update(additional_multiloc_map)

    return [
        ephys.locations.NrnSeclistLocation(sec, seclist_name=sec)
        for sec in multiloc_map.get(section_name, [section_name])
    ]


def define_distributions(distributions_definition, morphology=None):
    """Create a list of ParameterScaler from a the definition of channel distributions

    Args:
        distributions_definition (list): definitions of the distributions
    """
    hotspot_begin, hotspot_end = None, None
    if any(definition.name == "step" for definition in distributions_definition):
        hotspot_begin, hotspot_end = get_hotspot_location(morphology.morphology_path)
    distributions = collections.OrderedDict()

    for definition in distributions_definition:
        if definition.name == "uniform":
            distributions[definition.name] = NrnSegmentLinearScaler()
        elif definition.name == "step":
            distributions[definition.name] = NrnSegmentSomaDistanceStepScaler(
                name=definition.name,
                distribution=definition.function,
                dist_param_names=definition.parameters,
                soma_ref_location=definition.soma_ref_location,
                step_begin=hotspot_begin,
                step_end=hotspot_end,
            )
        else:
            distributions[definition.name] = NrnSegmentSomaDistanceScaler(
                name=definition.name,
                distribution=definition.function,
                dist_param_names=definition.parameters,
                soma_ref_location=definition.soma_ref_location,
            )

    return distributions


def define_parameters(parameters_definition, distributions, mapping_multilocation):
    """Define a list of NrnParameter from a definition dictionary

    Args:
        parameters_definition (list): definitions of the parameters
        distributions (list): list of distributions in the form of ParameterScaler
        mapping_multilocation (dict): mapping from multi-locations names to list of locations
    """

    parameters = []

    for param_def in parameters_definition:
        if isinstance(param_def.value, (list, tuple)):
            is_frozen = False
            value = None
            bounds = param_def.value
            if bounds[0] > bounds[1]:
                raise ValueError(
                    f"Lower bound ({bounds[0]}) is greater than upper bound ({bounds[1]})"
                    f" for parameter {param_def.name}"
                )
        else:
            is_frozen = True
            value = param_def.value
            bounds = None

        if param_def.location == "global":
            parameters.append(
                ephys.parameters.NrnGlobalParameter(
                    name=param_def.name,
                    param_name=param_def.name,
                    frozen=is_frozen,
                    bounds=bounds,
                    value=value,
                )
            )
            continue

        dist = None
        seclist_locations = None
        if "distribution_" in param_def.location:
            dist = distributions[param_def.location.split("distribution_")[1]]
        else:
            seclist_locations = multi_locations(param_def.location, mapping_multilocation)

        if dist:
            parameters.append(
                ephys.parameters.MetaParameter(
                    name=f"{param_def.name}.{param_def.location}",
                    obj=dist,
                    attr_name=param_def.name,
                    frozen=is_frozen,
                    bounds=bounds,
                    value=value,
                )
            )
        elif param_def.distribution != "uniform":
            parameters.append(
                ephys.parameters.NrnRangeParameter(
                    name=f"{param_def.name}.{param_def.location}",
                    param_name=param_def.name,
                    value_scaler=distributions[param_def.distribution],
                    value=value,
                    bounds=bounds,
                    frozen=is_frozen,
                    locations=seclist_locations,
                )
            )
        else:
            parameters.append(
                ephys.parameters.NrnSectionParameter(
                    name=f"{param_def.name}.{param_def.location}",
                    param_name=param_def.name,
                    value_scaler=distributions["uniform"],
                    value=value,
                    bounds=bounds,
                    frozen=is_frozen,
                    locations=seclist_locations,
                )
            )

    return parameters


def define_mechanisms(mechanisms_definition, mapping_multilocation):
    """Define a list of NrnMODMechanism from a definition dictionary

    Args:
        mechanisms_definition (list of MechanismConfiguration): definition of the mechanisms
        mapping_multilocation (dict): mapping from multi-locations names to list of locations

    Returns:
        list: list of NrnMODMechanism
    """

    mechanisms = []

    for mech_def in mechanisms_definition:
        seclist_locations = multi_locations(mech_def.location, mapping_multilocation)

        mechanisms.append(
            ephys.mechanisms.NrnMODMechanism(
                name=f"{mech_def.name}.{mech_def.location}",
                mod_path=None,
                prefix=mech_def.name,
                locations=seclist_locations,
                preloaded=True,
                deterministic=not mech_def.stochastic,
            )
        )

    return mechanisms


def define_morphology(
    model_configuration,
    do_set_nseg=True,
    nseg_frequency=40,
    morph_modifiers=None,
):
    """Define a morphology object from a morphology file

    Args:
        model_configuration (NeuronModelConfiguration): configuration of the model
        do_set_nseg (float): set the length for the discretization
            of the segments
        nseg_frequency (float): frequency of nseg
        morph_modifiers (list of str or list of list):
                If str, name of the morph modifier to use from bluepyemodel.evaluation.modifiers.
                If List of morphology modifiers. Each modifier is defined by a list
                that includes the following elements:
                1. The path to the file that contains the modifier.
                2. The name of the function that applies the modifier.
                3. Optionally, a "hoc_string" that represents the hoc code for the modifier.
                For example, morph_modifiers could be defined as follows:

                    .. code-block::

                        morph_modifiers = [["path_to_module",
                                            "name_of_function",
                                            "hoc_string"], ...].

                If the "hoc_string" is not provided, the system will search within
                the specified module for a string that matches the function name appended
                with "_hoc".
                If ``None``, the default modifier will replace the axon with a tappered axon
                initial segment (replace_axon_with_taper). If you do not wish to use any modifier,
                set the present argument to ``[]``.
                If ``["bluepyopt_replace_axon"]``, the replace_axon function from
                bluepyopt.ephys.morphologies.NrnFileMorphology will be used
                and no other morph modifiers will be used.

    Returns:
        bluepyopt.ephys.morphologies.NrnFileMorphology: a morphology object
    """
    do_replace_axon = False

    if isinstance(morph_modifiers, str):
        morph_modifiers = [morph_modifiers]

    if morph_modifiers is None or morph_modifiers == [None]:
        morph_modifiers = [replace_axon_with_taper]
        morph_modifiers_hoc = [replace_axon_hoc]  # TODO: check the hoc is correct
        logger.debug("No morphology modifiers provided, replace_axon_with_taper will be used.")
    elif morph_modifiers == ["bluepyopt_replace_axon"]:
        morph_modifiers = None
        morph_modifiers_hoc = None
        do_replace_axon = True
    else:
        morph_modifiers_hoc = [None] * len(morph_modifiers)
        for i, morph_modifier in enumerate(morph_modifiers):
            if isinstance(morph_modifier, list):
                if len(morph_modifier) < 2 or len(morph_modifier) > 3:
                    raise ValueError("Invalid morph_modifier structure.")

                try:
                    modifier_module = importlib.import_module(morph_modifier[0])
                except ModuleNotFoundError:
                    logger.warning("Module %s not found.", morph_modifier[0])
                    continue

                morph_modifiers[i] = _get_attribute(modifier_module, morph_modifier[1])
                if len(morph_modifier) == 3:
                    morph_modifiers_hoc[i] = _get_attribute(modifier_module, morph_modifier[2])
                else:
                    morph_modifiers_hoc[i] = _get_attribute(
                        modifier_module, morph_modifier[1], default_hoc=True
                    )

            elif isinstance(morph_modifier, str):
                if morph_modifier == "replace_axon_with_taper":
                    morph_modifiers[i] = replace_axon_with_taper
                    morph_modifiers_hoc[i] = replace_axon_hoc
                else:
                    morph_modifiers[i] = _get_attribute(modifiers, morph_modifier)
                    morph_modifiers_hoc[i] = _get_attribute(modifiers, morph_modifier + "_hoc")

            elif not callable(morph_modifier):
                raise TypeError(
                    "A morph modifier is not callable nor a string nor a list of two str"
                )

    return NrnFileMorphology(
        morphology_path=model_configuration.morphology.path,
        do_replace_axon=do_replace_axon,
        do_set_nseg=do_set_nseg,
        nseg_frequency=nseg_frequency,
        morph_modifiers=morph_modifiers,
        morph_modifiers_hoc=morph_modifiers_hoc,
    )


def create_cell_model(
    name,
    model_configuration,
    morph_modifiers=None,
    seclist_names=None,
    secarray_names=None,
    nseg_frequency=40,
):
    """Create a cell model based on a morphology, mechanisms and parameters

    Args:
        name (str): name of the model
        morphology (dict): morphology from emodel api .get_morphologies()
        model_configuration (NeuronModelConfiguration): Configuration of the neuron model,
            containing the parameters their locations and the associated mechanisms.
        morph_modifiers (list of str or list of list):
            If str, name of the morph modifier to use from bluepyemodel.evaluation.modifiers.
            If List of morphology modifiers. Each modifier is defined by a list
            that includes the following elements:
            1. The path to the file that contains the modifier.
            2. The name of the function that applies the modifier.
            3. Optionally, a "hoc_string" that represents the hoc code for the modifier.
            For example, morph_modifiers could be defined as follows:

                .. code-block::

                    morph_modifiers = [["path_to_module",
                                        "name_of_function",
                                        "hoc_string"], ...].

            If the "hoc_string" is not provided, the system will search within
            the specified module for a string that matches the function name appended
            with "_hoc".
            If ``None``, the default modifier will replace the axon with a tappered axon
            initial segment (replace_axon_with_taper). If you do not wish to use any modifier,
            set the present argument to ``[]``.
            If ``["bluepyopt_replace_axon"]``, the replace_axon function from
            bluepyopt.ephys.morphologies.NrnFileMorphology will be used
            and no other morph modifiers will be used.

    Returns:
        CellModel
    """

    morph = define_morphology(
        model_configuration,
        do_set_nseg=True,
        nseg_frequency=nseg_frequency,
        morph_modifiers=morph_modifiers,
    )

    if seclist_names is None:
        seclist_names = model_configuration.morphology.seclist_names
    if secarray_names is None:
        secarray_names = model_configuration.morphology.secarray_names

    mechanisms = define_mechanisms(
        model_configuration.mechanisms, model_configuration.mapping_multilocation
    )
    distributions = define_distributions(model_configuration.distributions, morph)
    parameters = define_parameters(
        model_configuration.parameters,
        distributions,
        model_configuration.mapping_multilocation,
    )

    return ephys.models.CellModel(
        name=name.replace(":", "_").replace("-", "_"),
        morph=morph,
        mechs=mechanisms,
        params=parameters,
        seclist_names=seclist_names,
        secarray_names=secarray_names,
    )


def _get_attribute(module, attribute_name, default_hoc=False):
    if module:
        attr = getattr(module, attribute_name, None)
        if attr is None and default_hoc:
            attr = getattr(module, attribute_name + "_hoc", None)
            attribute_name += "_hoc"
        if attr is None:
            logger.warning("%s not found in module %s.", attribute_name, module)
        return attr
    return None



"""EModelPipelineSettings class"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

# pylint: disable=too-many-arguments,too-many-locals,too-many-instance-attributes
# pylint: disable=too-many-statements

logger = logging.getLogger(__name__)


class EModelPipelineSettings:
    """Container for the settings used during the different steps of the e-model building pipeline.
    When using the "local" access point, these settings will be coming from the recipes.json file.
    When using the "nexus" access point, these settings will be coming from a resource of type
    EModelPipelineSettings.

    This is a backend container class, not meant to be used directly by the user."""

    def __init__(
        self,
        extraction_reader=None,
        extraction_threshold_value_save=1,
        plot_extraction=True,
        pickle_cells_extraction=False,
        extract_absolute_amplitudes=False,
        rheobase_strategy_extraction="absolute",
        rheobase_settings_extraction=None,
        interpolate_RMP_extraction=False,
        default_std_value=1e-3,
        bound_max_std=False,
        efel_settings=None,
        minimum_protocol_delay=0.0,
        stochasticity=False,
        morph_modifiers=None,
        threshold_based_evaluator=None,
        start_from_emodel=None,
        optimiser="IBEA",
        optimizer=None,
        optimisation_params=None,
        optimisation_timeout=600.0,
        threshold_efeature_std=None,
        max_ngen=100,
        optimisation_checkpoint_period=None,
        use_stagnation_criterion=True,
        validation_function="max_score",
        validation_threshold=5.0,
        plot_optimisation=True,
        use_ProbAMPANMDA_EMS=False,
        compile_mechanisms=False,
        n_model=3,
        optimisation_batch_size=5,
        max_n_batch=3,
        path_extract_config=None,
        name_Rin_protocol=None,
        name_rmp_protocol=None,
        validation_protocols=None,
        name_gene_map=None,
        plot_currentscape=True,
        plot_parameter_evolution=True,
        plot_bAP_EPSP=False,
        currentscape_config=None,
        save_recordings=False,
        neuron_dt=None,
        cvode_minstep=0.0,
        use_params_for_seed=True,
        max_threshold_voltage=-30,
        strict_holding_bounds=True,
        max_depth_holding_search=7,
        max_depth_threshold_search=10,
        spikecount_timeout=50,
        files_for_extraction=None,
        targets=None,
        protocols_rheobase=None,
        auto_targets=None,
        auto_targets_presets=None,
    ):
        """Creator of the EModelPipelineSettings class.

        Args:
            extraction_reader (function or list): function used to read the ephys data during
                efeature extraction. If it is a list, it must contain the path to the file
                containing the function and name of the function. E.g: ``["path_to_module",
                "name_of_function"]``. If None, BluePyEfe will try to automatically use the correct
                default reader.
            extraction_threshold_value_save (int): during extraction, define the minimum number of
                values (data points) needed for an e-feature to be considered valid and be returned
                in the output of the extraction process. If the number of data point is lower than
                this value, the e-feature will be ignored.
            plot_extraction (bool): at the end of extraction, should the e-features and
                experimental traces be plotted.
            pickle_cells_extraction (bool): at the end of extraction, should the BluePyEfe objects
                be saved in pickle files for further analysis. The pickle files will be
                saved in the folder "./figures/{emodel_name}/efeatures_extraction/".
            extract_absolute_amplitudes (bool): if True, during extraction, BluePyEfe will assume
                that the targets are expressed in absolute current amplitudes (nA) instead of the
                relative amplitudes when checking if a recording fits a given target.
            rheobase_strategy_extraction (str): during extraction, sets which function is used to
                compute the rheobase of the experimental cells. Can be 'absolute' (amplitude of
                the lowest current amplitude inducing at least a spike) or 'flush' (amplitude of
                the lowest current amplitude inducing at least a spike followed by another
                recording also inducing a spike).
            rheobase_settings_extraction (dict): settings related to the rheobase computation
                strategy. Keys have to match the arguments expected by the rheobase computation
                function present in the module bluepyefe.rheobase.
            interpolate_RMP_extraction (bool): whether to set the RMP after extraction as
                V_hold - R_in*I_Hold, which is an approximation of the RMP.
                This should be used when there is no protocol without holding current
                in the experimental data.
            default_std_value (float): At the end of e-features extraction, all features
                presenting a standard deviation of 0, will see their standard deviation
                replaced by the present value.
            bound_max_std (bool): If set to True, the std from extraction will be set to
                the mean value from extraction if it goes above it.
            efel_settings (dict): efel settings in the form {setting_name: setting_value} to be
                used during extraction. If settings are also informed in the targets on a per
                efeature basis, the latter will have priority.
            minimum_protocol_delay (float): during optimisation, if a protocol has an initial
                delay shorter than this value, the delay will be set to minimum_protocol_delay.
            stochasticity (bool or list of str): should channels behave stochastically if they can.
                If True, the mechanisms will be stochastic for all protocols. If a list of protocol
                names is provided, the mechanisms will be stochastic only for these protocols.
            morph_modifiers (list of str or list of list):
                If str, name of the morph modifier to use from bluepyemodel.evaluation.modifiers.
                If List of morphology modifiers. Each modifier is defined by a list
                that includes the following elements:
                1. The path to the file that contains the modifier.
                2. The name of the function that applies the modifier.
                3. Optionally, a "hoc_string" that represents the hoc code for the modifier.
                For example, morph_modifiers could be defined as follows:

                    .. code-block::

                        morph_modifiers = [["path_to_module",
                                            "name_of_function",
                                            "hoc_string"], ...].

                If the "hoc_string" is not provided, the system will search within
                the specified module for a string that matches the function name appended
                with "_hoc".
                If ``None``, the default modifier will replace the axon with a tappered axon
                initial segment (replace_axon_with_taper). If you do not wish to use any modifier,
                set the present argument to ``[]``.
                If ``["bluepyopt_replace_axon"]``, the replace_axon function from
                bluepyopt.ephys.morphologies.NrnFileMorphology will be used
                and no other morph modifiers will be used.
            threshold_based_evaluator (bool): not used. To be deprecated.
            start_from_emodel (dict): If informed, the optimisation for the present e-model will
                be instantiated using as values for the model parameters the ones from the
                e-model specified in the present dict. That option can be used for example
                to perform a two-steps optimisations. Example:

                .. code-block::

                        {
                            "emodel": "bNAC",
                            "etype": "bNAC",
                            "iteration_tag": "mytest"
                        }.
            optimiser (str): name of the algorithm used for optimisation, can be "IBEA", "SO-CMA"
                or "MO-CMA".
            optimizer (str): here for backward compatibility. Use optimiser instead.
            optimisation_params (dict): parameter for the optimisation process. Keys have to match
                the call of the optimiser. Here are the possible options and default values for the
                different optimisers:

                .. code-block::

                    - IBEA: {
                        "offspring_size": 100 # number of individuals in the population
                    }
                    - SO-CMA: {
                        "offspring_size": 20, # number of individuals in the population
                        sigma=0.4 # initial standard deviation of the gaussian distribution
                    }
                    - MO-CMA: {
                        "offspring_size": 20, # number of individuals in the population
                        sigma=0.4, # initial standard deviation of the gaussian distribution
                        weight_hv=0.5 # weight of the hypervolume score in the selection process
                    }

                For more details, see the documentation of the bluepyopt.deapext package.
            optimisation_timeout (float): duration (in second) after which the evaluation
                of a protocol will be interrupted. When a protocol is interrupted, its response
                will be considered as invalid and the score of all the related e-features will
                be set to the maximum value (250 per default).
            threshold_efeature_std (float): if informed, the standard deviations of the
                efeatures will be thresholded at a minimum of ``abs(threshold_efeature_std
                * efeature_mean)``. Note that this will not overwrite the original standard
                deviations, but only modify them for the optimisation process.
            max_ngen (int): maximum number of generations of the evolutionary process of the
                optimisation.
            optimisation_checkpoint_period (float): minimum time (in s) between checkpoint save.
                None to save checkpoint independently of the time between them
            use_stagnation_criterion (bool): whether to use the stagnation stopping criterion on
                top of the maximum generation criterion during optimisation.
            validation_function (str or list): if str, can be ``max_score`` or ``mean_score``.
                If list, must contain the path to the file containing the function and name
                of the function. E.g: ``["path_to_module", "name_of_function"]``
            validation_threshold (float): if ``max_score`` or ``mean_score`` were specified as the
                validation_function, this parameter will set the threshold under which the e-models
                will be considered to pass validation successfully.
            optimisation_batch_size (int): number of optimisation seeds to run in parallel. Only
                used by the Luigi workflow.
            max_n_batch (int): maximum number of optimisation batches. Only used by the Luigi
                workflow.
            name_gene_map (str): name of the gene mapping csv file. Only required when using the
                Nexus access_point and the IC-selector.
            n_model (int): minimum number of models to pass validation to consider the e-model
                building task done. Only used by the Luigi workflow.
            plot_optimisation (bool): should the e-models scores and traces be plotted. Only used
                by the Luigi workflow.
            use_ProbAMPANMDA_EMS (bool): True to link ProbAMPANMDA_EMS in EMC on nexus,
                and download ProbAMPANMDA from nexus along with other mechanisms.
            compile_mechanisms (bool): should the mod files be copied in the local
                mechanisms_dir directory. Only used by the Luigi workflow.
            path_extract_config (str): specify the path to the .json file containing the targets
                for the extraction process. Defaults to the current directory.
                If an 'iteration_tag' is provided, the file will be located in
                './run/iteration_tag'. This is only applicable with a local access point.
                See example emodel_pipeline_local_python for more details.
            name_Rin_protocol (list or str): name and amplitude of the protocol from which the
                input resistance should be selected from. The matching protocol should have
                "ohmic_input_resistance_vb_ssse" in its feature targets .E.g: ``["IV", -20]`` or
                ``IV_-20``.
                This setting has to be set before efeature extraction if you wish to run
                a threshold based evaluator.
            name_rmp_protocol (list or str): name and amplitude of the protocol from which the
                resting membrane potential should be selected from, e.g: ``["IV", 0]`` or ``IV_0``.
                The matching protocol should have "voltage_base" in its feature targets.
                This setting has to be set before efeature extraction if you wish to run
                a threshold based evaluator.
            validation_protocols (list of str): name of the protocols to be used for validation
                only. E.g. ``["APWaveform_300"]``. These protocols will not be used during
                optimisation.
            plot_currentscape (bool): during the plotting, should the currentscapes be
                plotted for the recordings.
            plot_parameter_evolution (bool): during the plotting, should the evolution of the
                parameters be plotted.
            plot_bAP_EPSP (bool): during the plotting, should ready-to-use back-propagating AP
                and EPSP protocols be run and plotted.
                Should be True only for pyramidal cells,
                since it depends on the presence of apical dendrite.
            currentscape_config (dict): currentscape configuration according to the currentscape
                documentation (https://github.com/BlueBrain/Currentscape).
                Note that current.names, output.savefig, output.fname and output.dir
                do not need to be set, since they are automatically overwritten by BluePyEModel.
                If current.names is set nonetheless, it will be used as the subset of available
                currents to be selected for the plot.
            save_recordings (bool): Whether to save the responses data under a folder
                named `recordings`.
            neuron_dt (float): time step of the NEURON simulator. If ``None``, cvode will be used.
            cvode_minstep (float): minimum time step allowed when using cvode.
            use_params_for_seed (bool): use a hashed version of the parameter
                dictionary as a seed for the simulator
            max_threshold_voltage (float): upper bound for the voltage during the
                search for the threshold or rheobase current (see SearchThresholdProtocol).
            strict_holding_bounds (bool): if True, the minimum and maximum values for the current
                used during the holding current search will be fixed. Otherwise, they will be
                widened dynamically if the holding current is beyond the initial bounds.
            max_depth_holding_search (int): maximum depth for the binary search for the
                holding current.
            max_depth_threshold_search (int): maximum depth for the binary search for the
                threshold current.
            spikecount_timeout (float): during the search of the threshold current, if the present
                timeout is reached, we set spikecount=2 as if many spikes were present, to speed
                up bisection search.
            files_for_extraction (list): temporary
            targets (list): temporary
            protocols_rheobase (list): temporary
            auto_targets (list): temporary
            auto_targets_presets (list): temporary
        """

        # Settings related to E-features extraction
        self.extraction_reader = extraction_reader
        self.extraction_threshold_value_save = extraction_threshold_value_save
        self.plot_extraction = plot_extraction
        self.extract_absolute_amplitudes = extract_absolute_amplitudes
        self.pickle_cells_extraction = pickle_cells_extraction
        self.efel_settings = efel_settings  # Also used during optimisation
        if self.efel_settings is None:
            self.efel_settings = {"interp_step": 0.025, "strict_stiminterval": True}
        self.path_extract_config = path_extract_config
        self.rheobase_strategy_extraction = rheobase_strategy_extraction
        self.rheobase_settings_extraction = rheobase_settings_extraction
        self.interpolate_RMP_extraction = interpolate_RMP_extraction
        self.default_std_value = default_std_value
        self.bound_max_std = bound_max_std
        self.minimum_protocol_delay = minimum_protocol_delay

        # Settings related to the evaluator
        self.max_threshold_voltage = max_threshold_voltage
        self.threshold_efeature_std = threshold_efeature_std
        self.max_depth_holding_search = max_depth_holding_search
        self.max_depth_threshold_search = max_depth_threshold_search
        self.spikecount_timeout = spikecount_timeout
        self.stochasticity = stochasticity
        self.neuron_dt = neuron_dt
        self.cvode_minstep = cvode_minstep
        self.use_params_for_seed = use_params_for_seed

        # Settings related to the optimisation
        self.start_from_emodel = start_from_emodel
        self.optimisation_timeout = optimisation_timeout
        self.optimiser = optimiser if optimizer is None else optimizer
        self.optimisation_params = optimisation_params
        if self.optimisation_params is None:
            if self.optimiser == "IBEA":
                self.optimisation_params = {"offspring_size": 100}
            else:
                self.optimisation_params = {"offspring_size": 20}

        self.max_ngen = max_ngen
        self.optimisation_checkpoint_period = optimisation_checkpoint_period
        self.use_stagnation_criterion = use_stagnation_criterion
        self.plot_optimisation = plot_optimisation
        self.compile_mechanisms = compile_mechanisms

        # Specific to threshold based optimisation
        self.name_Rin_protocol = name_Rin_protocol
        self.name_rmp_protocol = name_rmp_protocol
        self.strict_holding_bounds = strict_holding_bounds

        # Settings related to the validation
        self.validation_threshold = validation_threshold
        self.validation_function = validation_function

        if isinstance(validation_protocols, dict):
            self.validation_protocols = [
                f"{name}_{amp}"
                for name in validation_protocols
                for amp in validation_protocols[name]
            ]
        elif validation_protocols is None:
            self.validation_protocols = []
        else:
            self.validation_protocols = validation_protocols

        # Settings specific to the Luigi pipeline
        self.n_model = n_model
        self.optimisation_batch_size = optimisation_batch_size
        self.max_n_batch = max_n_batch
        self.name_gene_map = name_gene_map

        # Settings specific to the currentscape plotting
        self.plot_currentscape = plot_currentscape
        self.currentscape_config = currentscape_config

        self.plot_parameter_evolution = plot_parameter_evolution
        self.plot_bAP_EPSP = plot_bAP_EPSP

        # Settings specific to the recordings
        self.save_recordings = save_recordings

        # Settings specific to the cell model
        self.morph_modifiers = morph_modifiers

        if threshold_based_evaluator is not None:
            logger.warning(
                "Setting threshold_based_evaluator is not used anymore and will be deprecated"
            )

        # Settings for targets configuration
        # Temporarily in pipeline_settings - will come from SBO in the future
        self.files_for_extraction = files_for_extraction
        self.targets = targets
        self.protocols_rheobase = protocols_rheobase
        self.auto_targets = auto_targets
        self.auto_targets_presets = auto_targets_presets

        # Settings used with nexus
        self.use_ProbAMPANMDA_EMS = use_ProbAMPANMDA_EMS

    def as_dict(self):
        return vars(self)



"""Methods to handle cell model configuration."""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from dataclasses import asdict
from dataclasses import dataclass


@dataclass
class Parameter:
    """Used to exchange configuration information with BPEM."""

    name: str  # Neuron parameter name, i.e. {param}_{suffix}
    value: tuple  # Optimization boundaries
    location: str  # Cell compartment
    mechanism: str = ""  # Mechanism suffix
    distribution: str = ""  # Morphology dependent function


class Configuration:
    """Class to collect cell model parameters and generate a
    configuration."""

    def __init__(self):
        self._parameters = []

    @staticmethod
    def _clean_up_list_of_dicts(list_of_dicts, sort_by=None):
        """Sort list of dicts and remove duplicates.

        Args:
            sort_by (str): dict key to sort by
        """

        # Make unique list
        list_of_dicts = [dict(s) for s in set(frozenset(x.items()) for x in list_of_dicts)]
        # Sort by key
        if sort_by:
            list_of_dicts = sorted(list_of_dicts, key=lambda k: k[sort_by])
        return list_of_dicts

    def _add_mech_param(self, mech, param):
        """Add parameter from provided mechanism.

        Args:
            mech (Mechanism): mechanism containing parameter info
            param (str): name of parameter to add
        """

        model = mech.model
        suffix = model["suffix"]
        locations = asdict(mech.distribution)
        name = param + "_" + suffix
        for comp, distr in locations.items():
            if not distr == "":
                pset = Parameter(
                    name=name,
                    value=tuple(mech.get_bounds(param)),
                    location=comp,
                    mechanism=suffix,
                    distribution=distr,
                )
                self._parameters.append(pset)

    def add_from_mechanism(self, mech):
        """Add all parameters from provided mechanism.

        Args:
            mech (Mechanism): mechanism to add
        """

        model = mech.model

        if "gbar" in model:
            param = model["gbar"]
            self._add_mech_param(mech, param)

        if "parameters" in model:
            for param in model["parameters"]:
                self._add_mech_param(mech, param)

    def add_parameter(self, name, location, value):
        """Manually add a cell model parameter.

        Args:
            name (str): parameter name
            location (str): compartment to insert parameter
            value (tuple): bounds for parameter optimization
        """

        pset = Parameter(name=name, location=location, value=value)
        self._parameters.append(pset)

    def get_mechanisms(self):
        """Returns a list of mechanisms and compartments to insert them."""

        mechanisms = [
            {"name": p.mechanism, "location": p.location} for p in self._parameters if p.mechanism
        ]
        return self._clean_up_list_of_dicts(mechanisms, "name")

    def get_parameters(self):
        """Returns a list of configuration parameters."""

        params = [asdict(p) for p in self._parameters]
        return self._clean_up_list_of_dicts(params, "mechanism")

    def get_distributions(self):
        """Returns a list of unique distribution definitions used."""

        distr = [{"name": p.distribution} for p in self._parameters if p.distribution]
        return self._clean_up_list_of_dicts(distr, "name")

    def __str__(self):
        locations = {}
        for param in self._parameters:
            if param.location not in locations:
                locations[param.location] = []
            locations[param.location].append([param.name, param.value, param.distribution])
        out_str = ["\n"]
        for loc, content in locations.items():
            out_str.append(f">>> {loc} <<<")
            for p in content:
                if p[2]:
                    out_str.append(f"    {p[0]}, value: {p[1]}, distribution: {p[2]}")
                else:
                    out_str.append(f"    {p[0]}, value: {p[1]}")
        return "\n".join(out_str)



"""Mechanism Configuration"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from itertools import chain


class MechanismConfiguration:
    """Contains the information related to the definition and configuration of a mechanism"""

    def __init__(
        self,
        name,
        location,
        stochastic=None,
        version=None,
        temperature=None,
        ljp_corrected=None,
        parameters=None,
        ion_currents=None,
        nonspecific_currents=None,
        ionic_concentrations=None,
        id=None,
    ):
        """Init

        Args:
             name (str): name of the mechanism. Should be the SUFFIX from the mod file.
             locations (str or list of str): sections of the neuron on which this mechanism
                 will be instantiated.
             stochastic (bool): Can the mechanisms behave stochastically (optional).
             version (str): version id of the mod file.
             temperature (int): temperature of the mechanism
             ljp_corrected (bool): whether the mechanism is ljp corrected
             parameters (list): list of the possible parameter for this mechanism.
             ion_currents (list): list of the ion currents that this mechanism writes.
             nonspecific_currents (list): list of non-specific currents
             ionic_concentrations (list): list of the ionic concentration linked to the ion current
                If None, will be deduced from the ions list.
             id (str): Optional. Nexus ID of the mechanism.
        """

        self.name = name
        self.location = location
        self.version = version
        self.temperature = temperature
        self.ljp_corrected = ljp_corrected
        self.ion_currents = ion_currents
        self.nonspecific_currents = nonspecific_currents
        self.ionic_concentrations = ionic_concentrations
        if self.ionic_concentrations is None:
            self.ionic_concentrations = []
            if self.ion_currents is not None:
                for ion in self.ion_currents:
                    # remove 'i' in the front and put 'i' at the back to make it a concentration
                    self.ionic_concentrations.append(f"{ion[1:]}i")

        self.stochastic = stochastic
        if self.stochastic is None:
            self.stochastic = "Stoch" in self.name

        if parameters is None:
            self.parameters = {}
        elif isinstance(parameters, str):
            self.parameters = {parameters: [None, None]}
        else:
            self.parameters = parameters

        self.id = id

    def get_current(self):
        """Return the ion current names."""
        current = []
        ion_currents = self.ion_currents if self.ion_currents is not None else []
        nonspecific_currents = (
            self.nonspecific_currents if self.nonspecific_currents is not None else []
        )
        for curr in list(chain.from_iterable((ion_currents, nonspecific_currents))):
            current.append(f"{curr}_{self.name}")
        return current

    def as_dict(self):
        return {
            "name": self.name,
            "stochastic": self.stochastic,
            "location": self.location,
            "version": self.version,
            "temperature": self.temperature,
            "ljp_corrected": self.ljp_corrected,
            "id": self.id,
        }



"""Mechanism class corresponding to mechanisms fields in the icmapping file."""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from pprint import pprint

import numpy as np

from .distribution import Distribution
from .distribution import asdict


class Mechanism:
    """Holds all collected information associated with a Neuron mechanism.
    Corresponds to mechanisms fields in the icmapping file, with added
    fields containing additional information from GeneSelector."""

    def __init__(self, **kwargs):
        self.nexus = {}
        self.model = {}
        self.requires = []
        self.status = None
        self.distribution = Distribution()
        self._mapped_from = []
        self._bounds = {}
        self._selected = False
        for k, v in kwargs.items():
            setattr(self, k, v)
        # Some initial values
        if "parameters" in self.model:
            for param in self.model["parameters"]:
                self.set_parameters(**{param: 0})
        self.set_gbar([0, 1])

    def select(self, check_status=None):
        """Select mechanism for inclusion in model configuration.
        Args:
            check_status (str): only select the mechanism if status meets criterium
        """

        if not self.status:
            self._selected = True
            return

        if check_status == "stable":
            self._selected = check_status == self.status
        else:
            self._selected = True

    def deselect(self):
        """Deselect mechanism."""

        self._selected = False

    def is_selected(self):
        """Mechanism has been selected for inclusion in model configuration."""

        return self._selected

    def get_bounds(self, param):
        """Get parameter bounds.
        Args:
            param (str): parameter name

        Returns:
            bounds (list): parameter bounds
        """

        return self._bounds[param]

    def set_distribution(self, *args, **kwargs):
        """Set gbar distributions for all compartments.
        Args:
            args (list): distribution (str or Distribution)
            kwargs (dict): fields to be set specified as {compartment: distribution}
        """

        if len(args) > 0:
            if isinstance(args[0], Distribution):
                kwargs = asdict(args[0])
                self.distribution.set_fields(**kwargs)
            else:
                self.distribution.set_all(args[0])
        else:
            self.distribution.set_fields(**kwargs)

    def set_parameters(self, **kwargs):
        """Set the values of model parameters other than gbar
        Args:
            kwargs (dict): parameter values formatted as {parameter: value}
        """

        if "parameters" in self.model:
            for name, value in kwargs.items():
                if name in self.model["parameters"]:
                    if not isinstance(value, list):
                        value = [value]
                    self._bounds[name] = value
                else:
                    raise KeyError(f"Model '{self.model['suffix']}' has no parameter '{name}'")

    def set_gbar(self, value):
        """Set the value of gbar
        Args:
            value (list or number): value or bounds of gbar
        """

        if "gbar" in self.model:
            gbar = self.model["gbar"]
            if not isinstance(value, list):
                value = [value]
            value_check = [isinstance(v, (float, int)) for v in value]
            if not np.sum(value_check) == len(value):
                raise TypeError("gbar value must be numeric.")
            self._bounds[gbar] = value

    def set_from_gene_info(self, info):
        """Set fields based on gene info.
        Args:
            info (dict): channel information coming from the GeneSelector
        """

        self.set_distribution(info["distribution"])
        gbar = [0, info["gbar_max"]]
        self.set_gbar(gbar)
        self._mapped_from.append(f"{info['channel']}")

    def set_from_icmap(self, info):
        """Set fields based on icmap info.
        Args:
            info (dict): channel information coming from the icmapping file
        """

        if isinstance(info, list):
            self.set_gbar(info)
        elif isinstance(info, dict):
            for key, value in info.items():
                if key == "gbar":
                    self.set_gbar(value)
                elif key == "distribution":
                    if isinstance(value, str):
                        self.set_distribution(value)
                    elif isinstance(value, dict):
                        self.set_distribution(**value)
                elif key == "bounds":
                    self.set_parameters(**value)

    def asdict(self):
        """Return fields as a dict"""

        return self.__dict__

    def print(self):
        """Alternative print method"""

        pprint(self.asdict())

    def __str__(self):
        pstr = ", ".join([f"{k} = {v}" for k, v in self._bounds.items()])
        out_str = "{name}, distribution: {dist}, bounds: {param}"
        return out_str.format(name=self.model["suffix"], dist=self.distribution, param=pstr)



"""EModel_pipeline class."""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import glob
import logging
import pathlib

from bluepyemodel.access_point import get_access_point
from bluepyemodel.efeatures_extraction.efeatures_extraction import extract_save_features_protocols
from bluepyemodel.emodel_pipeline import plotting
from bluepyemodel.evaluation.evaluation import get_evaluator_from_access_point
from bluepyemodel.export_emodel.export_emodel import export_emodels_sonata
from bluepyemodel.model.model_configuration import configure_model
from bluepyemodel.optimisation import setup_and_run_optimisation
from bluepyemodel.optimisation import store_best_model
from bluepyemodel.tools.multiprocessing import get_mapper
from bluepyemodel.tools.utils import get_checkpoint_path
from bluepyemodel.validation.validation import validate

logger = logging.getLogger()


class EModel_pipeline:
    """The EModel_pipeline class is there to allow the execution of the steps
    of the e-model building pipeline using python (as opposed to the Luigi workflow).

    For an example of how to use to present class, see the example emodel_pipeline_local_python or
    the README.rst file."""

    def __init__(
        self,
        emodel,
        etype=None,
        ttype=None,
        mtype=None,
        species=None,
        brain_region=None,
        iteration_tag=None,
        morph_class=None,
        synapse_class=None,
        layer=None,
        recipes_path=None,
        use_ipyparallel=None,
        use_multiprocessing=None,
        data_access_point=None,
    ):
        """Initializes the EModel_pipeline.

        Args:
            emodel (str): name of the emodel. Can be arbitrary but has to match the name of the
                emodel in the recipes.json configuration file.
            etype (str): name of the e-type of the e-model. Used as an identifier for the e-model.
            ttype (str): name of the t-type of the e-model. Used as an identifier for the e-model.
                This argument is required when using the gene expression or IC selector.
            mtype (str): name of the m-type of the e-model. Used as an identifier for the e-model.
            species (str): name of the species of the e-model. Used as an identifier for the
                e-model.
            brain_region (str): name of the brain region of the e-model. Used as an identifier for
                the e-model.
            iteration_tag (str): tag associated to the current run. Used as an identifier for the
                e-model.
                If used with an access point of type "local", the current pipeline will execute
                the model building steps in the subdirectory of ``./run/{iteration_tag}/`` expected
                to contain a copy of the configuration files, mechanisms, morphologies needed for
                model building. This subdirectory can be created, for example using the following
                shell script (see also the example emodel_pipeline_local_python):

                .. code-block:: shell

                    git add -A && git commit --allow-empty -a -m "Running optimization"
                    export iteration_tag=$(git rev-parse --short HEAD)
                    git archive --format=tar --prefix=${iteration_tag}/ HEAD
                    | (cd ./run/ && tar xf -)

                In this case, the current, the iteration_tag can then be passed during the
                instantiation of the EModel_pipeline.
            morph_class (str): name of the morphology class, has to be "PYR", "INT". To be
                depracted.
            synapse_class (str): name of the synapse class of the e-model, has to be "EXC", "INH".
                Not used at the moment.
            layer (str): layer of the e-model. To be depracted.
            recipes_path (str): path of the recipes.json configuration file.This configuration
                file is the main file required when using the access point of type "local". It
                is expected to be a json file containing a dictionary whose keys are the names
                of the e-models that will be built. The values associated to these keys are
                the recipes used to build these e-models. See the example recipes.json file in
                the example emodel_pipeline_local_python for more details.
            use_ipyparallel (bool): should the parallelization map used for the different steps of
                the e-model building pipeline be based on ipyparallel.
            use_multiprocessing (bool): should the parallelization map used for the different steps
                of the e-model building pipeline be based on multiprocessing.
            data_access_point (str): Used for legacy purposes only
        """

        # pylint: disable=too-many-arguments

        if use_ipyparallel and use_multiprocessing:
            raise ValueError(
                "use_ipyparallel and use_multiprocessing cannot be both True at the same time. "
                "Please choose one."
            )
        if use_ipyparallel:
            self.mapper = get_mapper(backend="ipyparallel")
        elif use_multiprocessing:
            self.mapper = get_mapper(backend="multiprocessing")
        else:
            self.mapper = map

        endpoint = None

        if data_access_point is not None and data_access_point != "local":
            raise ValueError(
                "Attempted to set a legacy variable. "
                "This variable should not be modified in new code."
            )

        self.access_point = get_access_point(
            emodel=emodel,
            etype=etype,
            ttype=ttype,
            mtype=mtype,
            species=species,
            brain_region=brain_region,
            iteration_tag=iteration_tag,
            morph_class=morph_class,
            synapse_class=synapse_class,
            layer=layer,
            access_point="local",
            recipes_path=recipes_path,
            final_path="final.json",
            endpoint=endpoint,
        )

    def configure_model(
        self,
        morphology_name,
        morphology_path=None,
        morphology_format=None,
        use_gene_data=False,
    ):
        """To be deprecated"""

        return configure_model(
            self.access_point,
            morphology_name=morphology_name,
            morphology_path=morphology_path,
            morphology_format=morphology_format,
            use_gene_data=use_gene_data,
        )

    def extract_efeatures(self):
        """Extract the e-features related to the current e-model."""

        return extract_save_features_protocols(access_point=self.access_point, mapper=self.mapper)

    def optimise(self, seed=1):
        """Optimise the e-model.

        Args:
            seed (int): seed used for the random number generator used in the optimisation.
        """

        setup_and_run_optimisation(
            self.access_point,
            seed=seed,
            mapper=self.mapper,
            terminator=None,
        )

    def store_optimisation_results(self, seed=None):
        """Store the results of the optimisation. That is, reads the pickles file containing
        checkpoint of the optimisations and store the best model for each seed. When using the
        "local" access point, the best models are stored in a local json file called "final.json"
        When using the "nexus" access point, the best models are stored in Nexus resources of type
        EModel.

        Args:
            seed (int): specifies which seed to store. If None, the best models for all seeds
                are stored.
        """

        if seed is not None:
            store_best_model(access_point=self.access_point, seed=seed)

        else:
            checkpoint_path = get_checkpoint_path(self.access_point.emodel_metadata, seed=1)
            checkpoint_list = glob.glob(checkpoint_path.replace("seed=1", "*"))

            for chkp_path in checkpoint_list:
                file_name = pathlib.Path(chkp_path).stem
                tmp_seed = next(
                    int(e.replace("seed=", "")) for e in file_name.split("__") if "seed=" in e
                )

                store_best_model(
                    access_point=self.access_point,
                    seed=tmp_seed,
                    checkpoint_path=chkp_path,
                )

    def validation(self, preselect_for_validation=False):
        """Run a validation on the stored e-models. To work, some protocols have to be
        marked as for validation only. If no protocol is marked as such, the validation will
        simply check if the scores are all below a given threshold."""

        validate(
            access_point=self.access_point,
            mapper=self.mapper,
            preselect_for_validation=preselect_for_validation,
        )

    def plot(self, only_validated=False, load_from_local=False):
        """Plot the results of the optimisation in a subfolder called "figures".

        Args:
            only_validated (bool): if True, only the e-models that have successfully passed
                validation will be plotted.
            load_from_local (bool): if True, loads responses of the e-models from local files
                instead of recomputing them. Responses are automatically saved locally when
                plotting currentscapes.
        """

        cell_evaluator = get_evaluator_from_access_point(
            self.access_point,
            include_validation_protocols=True,
            record_ions_and_currents=self.access_point.pipeline_settings.plot_currentscape,
        )

        chkp_paths = glob.glob("./checkpoints/**/*.pkl", recursive=True)
        if not chkp_paths:
            raise ValueError("The checkpoints directory is empty, or there are no .pkl files.")

        # Filter the checkpoints to plot
        checkpoint_paths = []
        for chkp_path in chkp_paths:
            if self.access_point.emodel_metadata.emodel not in chkp_path.split("/"):
                continue
            if (
                self.access_point.emodel_metadata.iteration
                and self.access_point.emodel_metadata.iteration not in chkp_path.split("/")
            ):
                continue
            checkpoint_paths.append(chkp_path)

            stem = str(pathlib.Path(chkp_path).stem)
            seed = int(stem.rsplit("seed=", maxsplit=1)[-1])

            plotting.optimisation(
                optimiser=self.access_point.pipeline_settings.optimiser,
                emodel=self.access_point.emodel_metadata.emodel,
                iteration=self.access_point.emodel_metadata.iteration,
                seed=seed,
                checkpoint_path=chkp_path,
                figures_dir=pathlib.Path("./figures")
                / self.access_point.emodel_metadata.emodel
                / "optimisation",
            )

        if self.access_point.pipeline_settings.plot_parameter_evolution:
            plotting.evolution_parameters_density(
                evaluator=cell_evaluator,
                checkpoint_paths=checkpoint_paths,
                metadata=self.access_point.emodel_metadata,
                figures_dir=pathlib.Path("./figures")
                / self.access_point.emodel_metadata.emodel
                / "parameter_evolution",
            )

        return plotting.plot_models(
            access_point=self.access_point,
            mapper=self.mapper,
            seeds=None,
            figures_dir=pathlib.Path("./figures") / self.access_point.emodel_metadata.emodel,
            plot_distributions=True,
            plot_scores=True,
            plot_traces=True,
            plot_thumbnail=True,
            plot_currentscape=self.access_point.pipeline_settings.plot_currentscape,
            plot_bAP_EPSP=self.access_point.pipeline_settings.plot_bAP_EPSP,
            plot_dendritic_ISI_CV=True,
            plot_dendritic_rheobase=True,
            only_validated=only_validated,
            save_recordings=self.access_point.pipeline_settings.save_recordings,
            load_from_local=load_from_local,
            cell_evaluator=cell_evaluator,
        )

    def export_emodels(self, only_validated=False, seeds=None):
        """Export the e-models in the SONATA format. The results of the export are stored
        in a subfolder called export_emodels_sonata.

        Args:
            only_validated (bool): if True, only the e-models that have successfully passed
                validation will be exported.
            seeds (list): list of the seeds of the e-models to export. If None, all the e-models
                are exported.
        """

        export_emodels_sonata(
            self.access_point, only_validated, seeds=seeds, map_function=self.mapper
        )

    def summarize(self):
        """Prints a summary of the state of the current e-model building procedure"""

        print(self.access_point)


def sanitize_gitignore():
    """In order to avoid git issue when archiving the current working directory,
    adds the following lines to .gitignore: 'run/', 'checkpoints/', 'figures/',
    'logs/', '.ipython/', '.ipynb_checkpoints/'"""

    path_gitignore = pathlib.Path("./.gitignore")

    if not (path_gitignore.is_file()):
        raise FileNotFoundError("Could not update .gitignore as it does not exist.")

    with open(str(path_gitignore), "r") as f:
        lines = f.readlines()

    lines = " ".join(line for line in lines)

    to_add = [
        "run/",
        "checkpoints/",
        "figures/",
        "logs/",
        ".ipython/",
        ".ipynb_checkpoints/",
    ]

    with open(str(path_gitignore), "a") as f:
        for a in to_add:
            if a not in lines:
                f.write(f"{a}\n")



"""CustomFromFile luigi worker and custom luigi launcher."""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import signal

import luigi

from bluepyemodel.tasks.emodel_creation.optimisation import EModelCreation


class WorkerCustom(luigi.worker.Worker):
    """CustomFromFile Worker class."""

    def handle_interrupt(self, signum, _):
        """
        Trigger event to gracefully exit.
        Gracefully exit optimise deap loop, and allow to store models.
        All other pending tasks (optimise, validation, emodelcreation) are skipped.
        """
        if signum == signal.SIGUSR1:
            self._gracefully_exit()

    def _gracefully_exit(self):
        """
        Trigger event to gracefullt exit.
        Stop gracefully all Optimize tasks both running or pending.
        Pending Validation and EModelCreation are also skipped.
        """
        for task in self._scheduled_tasks.values():
            # trigger event all tasks
            if hasattr(task, "graceful_killer"):
                task.graceful_killer.set()

        # Force luigi not to check if deps are fulfilled
        # in case some Optimize tasks do not produce output
        # so that StoreBestModels still store best models of tasks that have run.
        self._config.check_unfulfilled_deps = False


class FactoryCustom(object):
    """CustomFromFile Worker Scheduler Factory class."""

    def create_local_scheduler(self):
        return luigi.scheduler.Scheduler(prune_on_get_work=True, record_task_history=False)

    def create_remote_scheduler(self, url):
        return luigi.rpc.RemoteScheduler(url)

    def create_worker(self, scheduler, worker_processes, assistant=False):
        # return your worker instance
        return WorkerCustom(
            scheduler=scheduler, worker_processes=worker_processes, assistant=assistant
        )


if __name__ == "__main__":
    luigi.build(
        [EModelCreation(emodel="L5PC", species="rat", seed=11)],
        worker_scheduler_factory=FactoryCustom(),
        local_scheduler=True,
    )



"""Efeatures extraction functions"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import pathlib
from importlib.machinery import SourceFileLoader

import bluepyefe.extract
import numpy

from bluepyemodel.evaluation.fitness_calculator_configuration import FitnessCalculatorConfiguration
from bluepyemodel.tools.search_pdfs import search_figure_efeatures

logger = logging.getLogger(__name__)


def interpolate_RMP(fitness_calculator_configuration):
    """If we do not have recordings with no holding, we need to estimate the RMP as:
    RMP = V_hold - R_in*I_Hold
    """
    rin = None
    holding_current = None
    holding_voltages = []

    for f in fitness_calculator_configuration.efeatures:
        if (
            f.efel_feature_name == "ohmic_input_resistance_vb_ssse"
            and f.protocol_name == "RinProtocol"
        ):
            rin = f.mean
        if f.efel_feature_name == "voltage_base" and (
            "iv" in f.protocol_name.lower()
            or "idrest" in f.protocol_name.lower()
            or "idthres" in f.protocol_name.lower()
        ):
            holding_voltages.append(f.mean)
        if f.efel_feature_name == "bpo_holding_current":
            holding_current = f.mean

    if rin is None:
        raise TypeError("Impossible to interpolate the RMP as Rin is also missing")
    if holding_current is None:
        raise TypeError("Impossible to interpolate the RMP as the holding current is also missing")
    if not holding_voltages:
        raise ValueError("Impossible to interpolate the RMP as no voltage base is available")

    holding_voltage = numpy.median(holding_voltages)
    rmp = holding_voltage - (rin * holding_current)

    logger.debug("The RMP was computed to be: %s", rmp)

    for i, f in enumerate(fitness_calculator_configuration.efeatures):
        if (
            f.efel_feature_name == "steady_state_voltage_stimend"
            and f.protocol_name == "RMPProtocol"
        ):
            fitness_calculator_configuration.efeatures[i].mean = rmp
            break
    else:
        raise ValueError("RMP feature not found and cannot be replaced")


def threshold_efeatures_std(fitness_calculator_configuration, default_std_value):
    """Set std as the mean value if std is higher than mean value."""
    for i, f in enumerate(fitness_calculator_configuration.efeatures):
        if f.original_std > abs(1.0 * f.mean) and f.original_std != default_std_value:
            logger.debug("Thresholding %s: %s -> %s", f.name, f.original_std, abs(1.0 * f.mean))
            fitness_calculator_configuration.efeatures[i].original_std = abs(1.0 * f.mean)


def define_extraction_reader_function(access_point):
    """Define the function used to read the ephys data during efeature extraction"""

    extraction_reader = access_point.pipeline_settings.extraction_reader

    if extraction_reader is None or not extraction_reader:
        logger.warning(
            "Extraction reader function not specified, BluePyEfe will use automatic "
            "detection based on file extension."
        )
        return None

    if isinstance(extraction_reader, list) and len(extraction_reader) == 2:
        # pylint: disable=deprecated-method,no-value-for-parameter
        function_module = SourceFileLoader(
            pathlib.Path(extraction_reader[0]).stem, extraction_reader[0]
        ).load_module()
        extraction_reader = getattr(function_module, extraction_reader[1])

    elif not callable(extraction_reader):
        raise TypeError("Extraction reader function is not callable nor a list of two strings")

    return extraction_reader


def attach_efeatures_pdf(emodel, efeatures):
    """If the efeatures are plotted, attach the path to the plot to the related efeature"""

    for protocol in efeatures:
        for efeat in efeatures[protocol]["soma"]:
            pdfs = {}

            pdf_amp, pdf_amp_rel = search_figure_efeatures(emodel, protocol, efeat["feature"])

            if pdf_amp:
                pdfs["amp"] = pdf_amp
            if pdf_amp_rel:
                pdfs["amp_rel"] = pdf_amp_rel

            if pdfs:
                efeat["pdfs"] = pdfs


def update_minimum_protocols_delay(access_point, config):
    """Threshold the minimum delay to the value of minimum_protocol_delay provided in the
    settings."""

    min_delay = access_point.pipeline_settings.minimum_protocol_delay

    for i, protocol in enumerate(config.protocols):
        if "delay" in protocol.stimuli[0] and protocol.stimuli[0]["delay"] < min_delay:
            logger.debug(
                "Replacing delay %s with %s in protocol %s",
                protocol.stimuli[0]["delay"],
                min_delay,
                protocol.name,
            )

            delta_delay = min_delay - protocol.stimuli[0]["delay"]
            config.protocols[i].stimuli[0]["delay"] = min_delay
            config.protocols[i].stimuli[0]["totduration"] = (
                protocol.stimuli[0]["totduration"] + delta_delay
            )

            for j, f in enumerate(config.efeatures):
                if f.protocol_name == protocol.name:
                    if "stim_start" in f.efel_settings:
                        config.efeatures[j].efel_settings["stim_start"] += delta_delay
                    if "stim_end" in f.efel_settings:
                        config.efeatures[j].efel_settings["stim_end"] += delta_delay

    return config


def extract_save_features_protocols(access_point, mapper=map):
    """Extract the efeatures and saves the results as a configuration for the fitness calculator.

    Args:
        access_point (DataAccessPoint): access point to the model's data
        mapper (map): mapper for parallel computations.
    """

    targets_configuration = access_point.get_targets_configuration()
    if (
        access_point.pipeline_settings.name_rmp_protocol is not None
        and access_point.pipeline_settings.name_Rin_protocol is not None
    ):
        targets_configuration.check_presence_RMP_Rin_efeatures(
            access_point.pipeline_settings.name_rmp_protocol,
            access_point.pipeline_settings.name_Rin_protocol,
        )

    reader_function = define_extraction_reader_function(access_point)

    threshold_nvalue_save = access_point.pipeline_settings.extraction_threshold_value_save
    plot = access_point.pipeline_settings.plot_extraction
    output_directory = f"./figures/{access_point.emodel_metadata.emodel}/efeatures_extraction/"

    efeatures, stimuli, current = bluepyefe.extract.extract_efeatures(
        output_directory=output_directory,
        files_metadata=targets_configuration.files_metadata_BPE,
        targets=targets_configuration.targets_BPE,
        auto_targets=targets_configuration.auto_targets_BPE,
        absolute_amplitude=access_point.pipeline_settings.extract_absolute_amplitudes,
        threshold_nvalue_save=threshold_nvalue_save,
        protocols_rheobase=targets_configuration.protocols_rheobase_BPE,
        recording_reader=reader_function,
        map_function=mapper,
        write_files=False,
        plot=plot,
        efel_settings=access_point.pipeline_settings.efel_settings,
        pickle_cells=access_point.pipeline_settings.pickle_cells_extraction,
        rheobase_strategy=access_point.pipeline_settings.rheobase_strategy_extraction,
        rheobase_settings=access_point.pipeline_settings.rheobase_settings_extraction,
        default_std_value=access_point.pipeline_settings.default_std_value,
    )

    if plot:
        attach_efeatures_pdf(access_point.emodel_metadata.emodel, efeatures)

    fitness_calculator_config = FitnessCalculatorConfiguration(
        name_rmp_protocol=access_point.pipeline_settings.name_rmp_protocol,
        name_rin_protocol=access_point.pipeline_settings.name_Rin_protocol,
        validation_protocols=access_point.pipeline_settings.validation_protocols,
        stochasticity=access_point.pipeline_settings.stochasticity,
        default_std_value=access_point.pipeline_settings.default_std_value,
    )

    fitness_calculator_config.init_from_bluepyefe(
        efeatures,
        stimuli,
        current,
        access_point.pipeline_settings.threshold_efeature_std,
        targets_configuration.protocols_mapping,
    )

    # correct for any weight present in features
    for target in targets_configuration.targets:
        if target.weight != 1.0:
            try:
                feat = next(
                    feat
                    for feat in fitness_calculator_config.efeatures
                    if feat.efel_feature_name == target.efeature
                    and feat.protocol_name == f"{target.protocol}_{target.amplitude}"
                )
                feat.weight = target.weight
            except StopIteration:
                logger.debug(
                    "Could not find efeature with name %s and protocol %s",
                    target.efeature,
                    target.protocol,
                )

    if access_point.pipeline_settings.bound_max_std:
        threshold_efeatures_std(
            fitness_calculator_config, access_point.pipeline_settings.default_std_value
        )

    fitness_calculator_config.protocols += fitness_calculator_config.initialise_protocols(
        targets_configuration.additional_fitness_protocols
    )

    fitness_calculator_config.efeatures += fitness_calculator_config.initialise_efeatures(
        targets_configuration.additional_fitness_efeatures,
        threshold_efeature_std=access_point.pipeline_settings.threshold_efeature_std,
    )

    if access_point.pipeline_settings.interpolate_RMP_extraction:
        interpolate_RMP(fitness_calculator_config)

    fitness_calculator_config = update_minimum_protocols_delay(
        access_point, fitness_calculator_config
    )

    access_point.store_fitness_calculator_configuration(fitness_calculator_config)

    return fitness_calculator_config



"""CustomFromFile stimulus class"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

import numpy

from bluepyemodel.ecode.stimulus import BPEM_stimulus

logger = logging.getLogger(__name__)


class CustomFromFile(BPEM_stimulus):
    """CustomFromFile current stimulus to be loaded from a file"""

    name = "CustomFromFile"

    def __init__(self, location, **kwargs):
        """Constructor

        Args:
            location(Location): location of stimulus
            **kwargs: See below

        Keyword Arguments:
            data_filepath (str): path to the noise .txt data file. The file should have two columns:
                time (ms) and current (nA).
        """

        data_filepath = kwargs["data_filepath"]
        series = numpy.loadtxt(data_filepath)

        self.time_series = series[:, 0]
        self.current_series = series[:, 1]

        super().__init__(
            location=location,
        )

    @property
    def total_duration(self):
        return self.time_series[-1]

    @property
    def stim_start(self):
        return 0.0

    @property
    def stim_end(self):
        return self.time_series[-1]

    def generate(self, dt=0.1):
        """Return current time series"""

        return self.time_series, self.current_series
